from __future__ import annotations
from typing import TYPE_CHECKING
import angr.ailment as ailment

from angr.analyses.decompiler.goto_manager import GotoManager
from angr.analyses import AnalysesHub
from angr.analyses.analysis import Analysis
from angr.analyses.decompiler.empty_node_remover import EmptyNodeRemover
from angr.analyses.decompiler.jump_target_collector import JumpTargetCollector
from angr.analyses.decompiler.redundant_label_remover import RedundantLabelRemover
from angr.analyses.decompiler.structuring.structurer_nodes import LoopNode
from angr.analyses.decompiler.semantic_naming.region_loop_counter_naming import RegionLoopCounterNaming
from .goto import GotoSimplifier
from .if_ import IfSimplifier
from .cascading_ifs import CascadingIfsRemover
from .ifelse import IfElseFlattener
from .loop import LoopSimplifier
from .expr_folding import (
    ExpressionCounter,
    ExpressionFolder,
    StoreStatementFinder,
    ExpressionLocation,
    InterferenceChecker,
    LoopNodeFinder,
)
from .cascading_cond_transformer import CascadingConditionTransformer
from .switch_expr_simplifier import SwitchExpressionSimplifier
from .switch_cluster_simplifier import SwitchClusterFinder, simplify_switch_clusters, simplify_lowered_switches

if TYPE_CHECKING:
    from angr.knowledge_plugins.variables.variable_manager import VariableManagerInternal


class RegionSimplifier(Analysis):
    """
    Simplifies a given region.
    """

    def __init__(
        self,
        func,
        region,
        arg_vvars: set[int] | None = None,
        simplify_switches: bool = True,
        simplify_ifelse: bool = True,
        variable_manager: VariableManagerInternal | None = None,
        loopctr_naming: bool = True,
    ):
        self.func = func
        self.region = region
        self.arg_vvars = arg_vvars
        self._simplify_switches = simplify_switches
        self._should_simplify_ifelses = simplify_ifelse
        self._variable_manager = variable_manager
        self._apply_loop_counter_naming = loopctr_naming

        self.goto_manager: GotoManager | None = None
        self.result = self.region

        self._simplify()

    def _simplify(self):
        """
        RegionSimplifier performs the following simplifications:
        - Remove redundant Gotos
        - Remove redundant If/If-else statements
        """

        r = self.region
        # Remove empty nodes
        r = self._remove_empty_nodes(r)
        # Remove unnecessary Jump statements
        r = self._simplify_gotos(r)
        # Remove unnecessary jump or conditional jump statements if they jump to the successor right afterwards
        r = self._simplify_ifs(r)
        # Remove labels that are not referenced by anything
        r = self._simplify_labels(r)
        # Remove empty nodes again
        r = self._remove_empty_nodes(r)

        if self.arg_vvars is not None:
            # Fold expressions that are only used once into their use sites
            r = self._fold_oneuse_expressions(r)
            r = self._remove_empty_nodes(r)

        if self._simplify_switches:
            # Simplify switch expressions
            r = self._simplify_switch_expressions(r)
            # Simplify switch clusters
            r = self._simplify_switch_clusters(r)
            # Again, remove labels that are not referenced by anything
            r = self._simplify_labels(r)

        # Remove empty nodes
        r = self._remove_empty_nodes(r)
        # Remove unnecessary else branches if the if branch will always return
        if self._should_simplify_ifelses:
            r = self._simplify_ifelses(r)
        #
        r = self._simplify_cascading_ifs(r)
        #
        r = self._simplify_loops(r)
        # Apply loop counter naming after loop simplification (when iterators are identified)
        if self._apply_loop_counter_naming and self._variable_manager is not None:
            self._apply_region_loop_counter_naming(r)
        # Remove empty nodes again
        r = self._remove_empty_nodes(r)
        # Find nested if-else constructs and convert them into CascadingIfs
        r = self._transform_to_cascading_ifs(r)

        self.result = r

    #
    # Simplifiers
    #

    def _fold_oneuse_expressions(self, region):
        loop_nodes = LoopNodeFinder(region).loop_nodes
        for sub_region in [*loop_nodes, region]:
            # fold one-use expressions in each sub-region
            if isinstance(sub_region, LoopNode):
                self._fold_oneuse_expressions_in_region(sub_region.sequence_node)
            else:
                self._fold_oneuse_expressions_in_region(sub_region)
        return region

    def _fold_oneuse_expressions_in_region(self, region):
        # pylint:disable=unreachable
        expr_counter = ExpressionCounter(region)

        variable_assignments = {}
        variable_uses = {}
        variable_assignment_dependencies = {}

        # pre-process and identify folding candidates
        # for variable definitions with loads, we invoke StoreStatementFinder to see if there are any Store statements
        # before the definition site and the use site.
        var_with_loads = {}
        single_use_variables = []
        for var, outerscope_uses in expr_counter.outerscope_uses.items():
            all_uses = expr_counter.all_uses[var]
            if (
                len(outerscope_uses) == 1
                and len(all_uses) == 1
                and var in expr_counter.assignments
                and len(expr_counter.assignments[var]) == 1
            ):
                definition, deps, loc, has_loads = next(iter(expr_counter.assignments[var]))
                _, use_expr_loc = next(iter(outerscope_uses))
                if isinstance(use_expr_loc, ExpressionLocation) and use_expr_loc.phi_stmt:
                    # we cannot fold expressions that are used in phi statements
                    continue
                if has_loads:
                    # the definition has at least one load expression. we need to ensure there are no store statements
                    # between the definition site and the use site
                    if isinstance(use_expr_loc, ExpressionLocation):
                        use_loc = use_expr_loc.statement_location()
                    else:
                        use_loc = use_expr_loc
                    var_with_loads[var] = (loc, use_loc)
                else:
                    single_use_variables.append(var)

        store_finder = StoreStatementFinder(region, set(var_with_loads.values()))
        for var, tpl in var_with_loads.items():
            if not store_finder.has_store(*tpl):
                single_use_variables.append(var)

        for var in single_use_variables:
            definition, deps, loc, _ = next(iter(expr_counter.assignments[var]))

            # make sure all variables that var depends on has been assigned at most once
            fail = False
            for dep_var in deps:
                if self.arg_vvars is not None and dep_var in self.arg_vvars:
                    continue
                if dep_var in expr_counter.assignments and len(expr_counter.assignments[dep_var]) > 1:
                    fail = True
                    break
            if fail:
                continue

            if isinstance(definition, ailment.Stmt.SideEffectStatement):
                # clear the existing variable since we no longer write to this variable after expression folding
                definition = definition.copy()
                if definition.ret_expr is not None:
                    definition.ret_expr = definition.ret_expr.copy()
                    definition.ret_expr.variable = None
            variable_assignments[var] = definition, loc
            variable_uses[var] = next(iter(expr_counter.outerscope_uses[var]))
            variable_assignment_dependencies[var] = deps

        # any variable definition that uses an existing to-be-removed variable cannot be folded
        all_variables_to_fold = set(variable_assignments)
        for var in all_variables_to_fold:
            if all_variables_to_fold.intersection(variable_assignment_dependencies[var]):
                del variable_assignments[var]
                del variable_uses[var]

        # ensure there is no interference between the call site and the use site
        checker = InterferenceChecker(variable_assignments, variable_uses, region)
        for varid in checker.interfered_assignments:
            if varid in variable_assignments:
                del variable_assignments[varid]
                del variable_uses[varid]
        # fold these expressions if possible
        ExpressionFolder(variable_assignments, variable_uses, region)
        return region

    @staticmethod
    def _simplify_switch_expressions(region):
        SwitchExpressionSimplifier(region)
        return region

    def _simplify_switch_clusters(self, region):
        finder = SwitchClusterFinder(region)
        simplify_switch_clusters(region, finder.var2condnodes, finder.var2switches)
        simplify_lowered_switches(
            region,
            {var: v for var, v in finder.var2condnodes.items() if var not in finder.var2switches},
            self.kb.functions,
        )
        return region

    @staticmethod
    def _remove_empty_nodes(region):
        return EmptyNodeRemover(region, claripy_ast_conditions=False).result

    @staticmethod
    def _transform_to_cascading_ifs(region):
        CascadingConditionTransformer(region)
        return region

    def _simplify_gotos(self, region):
        simplifier = GotoSimplifier(region, function=self.func, kb=self.kb)
        self.goto_manager = GotoManager(self.func, gotos=simplifier.irreducible_gotos)
        return region

    @staticmethod
    def _simplify_ifs(region):
        IfSimplifier(region)
        return region

    @staticmethod
    def _simplify_labels(region):
        jcl = JumpTargetCollector(region)
        RedundantLabelRemover(region, jcl.jump_targets)
        return region

    def _simplify_ifelses(self, region):
        IfElseFlattener(region, self.kb.functions)
        return region

    @staticmethod
    def _simplify_cascading_ifs(region):
        CascadingIfsRemover(region)
        return region

    def _simplify_loops(self, region):
        LoopSimplifier(region, self.kb.functions)
        return region

    def _apply_region_loop_counter_naming(self, region) -> None:
        assert self._variable_manager is not None
        namer = RegionLoopCounterNaming(region, self._variable_manager, self.kb.functions)
        namer.analyze()
        namer.apply_names()


AnalysesHub.register_default("RegionSimplifier", RegionSimplifier)



from __future__ import annotations
import logging
from typing import TYPE_CHECKING

from archinfo.arch_soot import SootAddressDescriptor

from angr.errors import SimValueError, SimSolverModeError
from angr.state_plugins.callstack import CallStack

if TYPE_CHECKING:
    from angr.sim_state import SimState
l = logging.getLogger(name=__name__)

# TODO: Make callsite an object and use it in BlockID and FunctionKey


class BlockID:
    """
    A context-sensitive key for a SimRun object.
    """

    def __init__(self, addr, callsite_tuples, jump_type):
        self.addr = addr
        self.callsite_tuples = callsite_tuples
        self.jump_type = jump_type

        self._hash = None

    def callsite_repr(self):
        if self.callsite_tuples is None:
            return "None"

        s = []

        def format_addr(addr):
            return "None" if addr is None else hex(addr)

        for i in range(0, len(self.callsite_tuples), 2):
            s.append("@".join(map(format_addr, self.callsite_tuples[i : i + 2])))
        return " -> ".join(s)

    def __repr__(self):
        return f"<BlockID {self.addr:#08x} ({self.callsite_repr()}) % {self.jump_type}>"

    def __hash__(self):
        if self._hash is None:
            self._hash = hash((self.callsite_tuples, self.addr, self.jump_type))
        return self._hash

    def __eq__(self, other):
        return (
            isinstance(other, BlockID)
            and self.addr == other.addr
            and self.callsite_tuples == other.callsite_tuples
            and self.jump_type == other.jump_type
        )

    def __ne__(self, other):
        return not self == other

    @staticmethod
    def new(addr, callstack_suffix, jumpkind):
        if jumpkind.startswith("Ijk_Sys") or jumpkind == "syscall":
            jump_type = "syscall"
        elif jumpkind in ("Ijk_Exit", "exit"):
            jump_type = "exit"
        else:
            jump_type = "normal"
        return BlockID(addr, callstack_suffix, jump_type)

    @property
    def func_addr(self):
        if self.callsite_tuples:
            return self.callsite_tuples[-1]
        return None


class FunctionKey:
    """
    A context-sensitive key for a function.
    """

    def __init__(self, addr, callsite_tuples):
        self.addr = addr
        self.callsite_tuples = callsite_tuples

        self._hash = None

    def callsite_repr(self):
        if self.callsite_tuples is None:
            return "None"

        s = []

        def format_addr(addr):
            return "None" if addr is None else hex(addr)

        for i in range(0, len(self.callsite_tuples), 2):
            s.append("@".join(map(format_addr, self.callsite_tuples[i : i + 2])))
        return " -> ".join(s)

    def __repr__(self):
        return f"<FuncKey {self.addr:#08x} ({self.callsite_repr()})>"

    def __hash__(self):
        if self._hash is None:
            self._hash = hash((self.callsite_tuples, self.addr))
        return self._hash

    def __eq__(self, other):
        return (
            isinstance(other, FunctionKey) and self.addr == other.addr and self.callsite_tuples == other.callsite_tuples
        )

    @staticmethod
    def new(addr, callsite_tuples):
        return FunctionKey(addr, callsite_tuples)


class CFGJobBase:
    """
    Describes an entry in CFG or VFG. Only used internally by the analysis.
    """

    def __init__(
        self,
        addr,
        state: SimState,
        context_sensitivity_level,
        block_id=None,
        src_block_id=None,
        src_exit_stmt_idx=None,
        src_ins_addr=None,
        jumpkind: str | None = None,
        call_stack=None,
        is_narrowing=False,
        skip=False,
        final_return_address=None,
    ):
        self.addr = addr  # Note that addr may not always be equal to self.state.ip (for syscalls, for example)
        self.state = state
        self.jumpkind = jumpkind
        self.src_block_id = src_block_id
        self.src_exit_stmt_idx = src_exit_stmt_idx
        self.src_ins_addr = src_ins_addr
        self.skip = skip
        self._block_id = block_id

        # Other parameters
        self._context_sensitivity_level = context_sensitivity_level
        self.is_narrowing = is_narrowing

        if call_stack is None:
            self._call_stack = CallStack()

            # Added the function address of the current exit to callstack
            se = self.state.solver
            sp_expr = self.state.regs.sp

            # If the sp_expr cannot be concretized, the stack pointer cannot be traced anymore.
            try:
                sp = se.eval_one(sp_expr)
            except (SimValueError, SimSolverModeError):
                l.warning("Stack pointer cannot be concretized. CallStack cannot track the stack pointer changes.")

                # Set the stack pointer to None
                sp = None

            self._call_stack = self._call_stack.call(
                None, self.addr, retn_target=final_return_address, stack_pointer=sp
            )

        else:
            self._call_stack = call_stack

    @property
    def call_stack(self):
        return self._call_stack

    def call_stack_copy(self):
        return self._call_stack.copy()

    def get_call_stack_suffix(self):
        return self._call_stack.stack_suffix(self._context_sensitivity_level)

    @property
    def func_addr(self):
        return self._call_stack.current_function_address

    @func_addr.setter
    def func_addr(self, v):
        # Make a copy because we might be sharing it with other CFGJobs
        self._call_stack = self._call_stack.copy()
        self._call_stack.current_function_address = v

    @property
    def current_stack_pointer(self):
        return self._call_stack.current_stack_pointer

    def __repr__(self):
        if isinstance(self.addr, SootAddressDescriptor):
            return f"<Entry {self.addr} {self.jumpkind}>"
        return f"<Entry {self.addr:#08x} % {self.jumpkind}>"



from __future__ import annotations
import itertools
from typing import TYPE_CHECKING

from angr.sim_type import (
    SimType,
    SimTypeChar,
    SimTypeShort,
    SimTypeInt,
    SimTypeLong,
    SimTypeLongLong,
    SimTypePointer,
    SimStruct,
    SimTypeArray,
    SimTypeFloat,
    SimTypeDouble,
    SimCppClass,
    SimTypeEnum,
    SimTypeWideChar,
    SimTypeFd,
)
from .typeconsts import (
    BottomType,
    Int8,
    Int16,
    Int32,
    Int64,
    Pointer32,
    Pointer64,
    Struct,
    Array,
    Float32,
    Float64,
    Enum,
    Fd,
)

if TYPE_CHECKING:
    from .typeconsts import TypeConstant


class TypeLifter:
    """
    Lift SimTypes to type constants.
    """

    __slots__ = ("bits", "memo", "named_struct_id_counter", "struct_name_to_idx")

    def __init__(self, bits: int):
        if bits not in (32, 64):
            raise ValueError("TypeLifter only supports 32-bit or 64-bit pointers.")
        self.bits = bits
        self.memo = {}
        self.named_struct_id_counter = itertools.count(133337)
        self.struct_name_to_idx = {}

    def lift(self, ty: SimType):
        handler = _mapping.get(type(ty))
        if handler is None:
            return BottomType(name=ty.label)

        return handler(self, ty)

    def _lift_SimTypeChar(self, ty):  # pylint:disable=no-self-use
        return Int8(name=ty.label)

    def _lift_SimTypeWideChar(self, ty):  # pylint:disable=no-self-use
        return Int16(name=ty.label)

    def _lift_SimTypeShort(self, ty):  # pylint:disable=no-self-use
        return Int16(name=ty.label)

    def _lift_SimTypeInt(self, ty):  # pylint:disable=no-self-use
        return Int32(name=ty.label)

    def _lift_SimTypeLongLong(self, ty):  # pylint:disable=no-self-use
        return Int64(name=ty.label)

    def _lift_SimTypePointer(self, ty: SimTypePointer):
        if self.bits == 32:
            return Pointer32(self.lift(ty.pts_to), name=ty.label)
        if self.bits == 64:
            return Pointer64(self.lift(ty.pts_to), name=ty.label)
        raise ValueError(f"Unsupported bits {self.bits}.")

    def _lift_SimStruct(self, ty: SimStruct) -> TypeConstant | BottomType:
        if ty in self.memo:
            return BottomType()

        struct_idx = {}
        if ty.name:
            if ty.name not in self.struct_name_to_idx:
                self.struct_name_to_idx[ty.name] = next(self.named_struct_id_counter)
            struct_idx["idx"] = self.struct_name_to_idx[ty.name]

        obj = Struct(fields={}, name=ty.name, **struct_idx)
        self.memo[ty] = obj
        converted_fields = {}
        field_names = {}
        ty_offsets = ty.offsets
        for field_name, simtype in ty.fields.items():
            if field_name not in ty_offsets:
                return BottomType()
            converted_fields[ty_offsets[field_name]] = self.lift(simtype)
            field_names[ty_offsets[field_name]] = field_name
        obj.fields = converted_fields
        obj.field_names = field_names
        del self.memo[ty]
        return obj

    def _lift_SimCppClass(self, ty: SimCppClass) -> TypeConstant | BottomType:
        if ty in self.memo:
            return BottomType()

        obj = Struct(fields={}, name=ty.name, is_cppclass=True)
        self.memo[ty] = obj
        converted_fields = {}
        field_names = {}
        ty_offsets = ty.offsets
        for field_name, simtype in ty.members.items():
            if field_name not in ty_offsets:
                return BottomType()
            converted_fields[ty_offsets[field_name]] = self.lift(simtype)
            field_names[ty_offsets[field_name]] = field_name
        obj.fields = converted_fields
        obj.field_names = field_names
        del self.memo[ty]
        return obj

    def _lift_SimTypeArray(self, ty: SimTypeArray) -> Array:
        elem_type = self.lift(ty.elem_type)
        return Array(elem_type, count=ty.length, name=ty.label)

    def _lift_SimTypeFloat(self, ty: SimTypeFloat) -> Float32:  # pylint:disable=no-self-use
        return Float32(name=ty.label)

    def _lift_SimTypeDouble(self, ty: SimTypeDouble) -> Float64:  # pylint:disable=no-self-use
        return Float64(name=ty.label)

    def _lift_SimTypeEnum(self, ty: SimTypeEnum) -> Enum:
        """Lift SimTypeEnum to Enum type constant."""
        base_tc = None
        if ty._base_type is not None:
            base_tc = self.lift(ty._base_type)
        return Enum(
            members=dict(ty.members),
            base_type=base_tc,
            name=ty.name,
        )

    def _lift_SimTypeFd(self, ty: SimTypeFd) -> Fd:  # pylint:disable=no-self-use
        return Fd(name=ty.label)


_mapping = {
    SimTypeChar: TypeLifter._lift_SimTypeChar,
    SimTypeWideChar: TypeLifter._lift_SimTypeWideChar,
    SimTypeShort: TypeLifter._lift_SimTypeShort,
    SimTypeInt: TypeLifter._lift_SimTypeInt,
    SimTypeLong: TypeLifter._lift_SimTypeInt,
    SimTypeLongLong: TypeLifter._lift_SimTypeLongLong,
    SimTypePointer: TypeLifter._lift_SimTypePointer,
    SimStruct: TypeLifter._lift_SimStruct,
    SimCppClass: TypeLifter._lift_SimCppClass,
    SimTypeArray: TypeLifter._lift_SimTypeArray,
    SimTypeFloat: TypeLifter._lift_SimTypeFloat,
    SimTypeDouble: TypeLifter._lift_SimTypeDouble,
    SimTypeEnum: TypeLifter._lift_SimTypeEnum,
    SimTypeFd: TypeLifter._lift_SimTypeFd,
}



# pylint:disable=isinstance-second-argument-not-valid-type,no-self-use,arguments-renamed,too-many-boolean-expressions
from __future__ import annotations
from collections.abc import Iterable
from abc import ABC, abstractmethod

import claripy
from typing_extensions import Self

from angr import ailment
from .utils import stable_hash, is_none_or_likeable, is_none_or_matchable
from .tagged_object import TaggedObject
from .expression import Atom, Expression, DirtyExpression


class Statement(TaggedObject, ABC):
    """
    The base class of all AIL statements.
    """

    __slots__ = ()

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def replace(self, old_expr: Expression, new_expr: Expression) -> tuple[bool, Self]:
        raise NotImplementedError

    def eq(self, expr0, expr1):  # pylint:disable=no-self-use
        if isinstance(expr0, claripy.ast.Base) or isinstance(expr1, claripy.ast.Base):
            return expr0 is expr1
        return expr0 == expr1

    @abstractmethod
    def likes(self, other) -> bool:  # pylint:disable=unused-argument,no-self-use
        raise NotImplementedError

    @abstractmethod
    def matches(self, other) -> bool:  # pylint:disable=unused-argument,no-self-use
        raise NotImplementedError

    def __eq__(self, other):
        if self is other:
            return True
        return type(self) is type(other) and self.likes(other) and self.idx == other.idx

    __hash__ = TaggedObject.__hash__


class Assignment(Statement):
    """
    Assignment statement: expr_a = expr_b
    """

    __slots__ = (
        "dst",
        "src",
    )

    def __init__(self, idx: int | None, dst: Atom, src: Expression, **kwargs):
        super().__init__(idx, **kwargs)

        self.dst = dst
        self.src = src

    def likes(self, other):
        return type(other) is Assignment and self.dst.likes(other.dst) and self.src.likes(other.src)

    def matches(self, other):
        return type(other) is Assignment and self.dst.matches(other.dst) and self.src.matches(other.src)

    def _hash_core(self):
        return stable_hash((Assignment, self.idx, self.dst, self.src))

    def __repr__(self):
        return f"Assignment ({self.dst}, {self.src})"

    def __str__(self):
        return f"{self.dst!s} = {self.src!s}"

    def replace(self, old_expr: Expression, new_expr: Expression):
        if self.dst == old_expr:
            r_dst = True
            assert isinstance(new_expr, Atom)
            replaced_dst = new_expr
        else:
            r_dst, replaced_dst = self.dst.replace(old_expr, new_expr)

        if self.src == old_expr:
            r_src = True
            replaced_src = new_expr
        else:
            r_src, replaced_src = self.src.replace(old_expr, new_expr)

        if r_dst or r_src:
            return True, Assignment(self.idx, replaced_dst, replaced_src, **self.tags)
        return False, self

    def copy(self) -> Assignment:
        return Assignment(self.idx, self.dst, self.src, **self.tags)

    def deep_copy(self, manager) -> Assignment:
        return Assignment(manager.next_atom(), self.dst.deep_copy(manager), self.src.deep_copy(manager), **self.tags)


class WeakAssignment(Statement):
    """
    An assignment statement that does not create a new variable at its destination; It should be seen as
    operator=(&dst, &src) in C++-like syntax.
    """

    __slots__ = (
        "dst",
        "src",
    )

    def __init__(self, idx: int | None, dst: Atom, src: Expression, **kwargs):
        super().__init__(idx, **kwargs)

        self.dst = dst
        self.src = src

    def likes(self, other):
        return type(other) is WeakAssignment and self.dst.likes(other.dst) and self.src.likes(other.src)

    def matches(self, other):
        return type(other) is WeakAssignment and self.dst.matches(other.dst) and self.src.matches(other.src)

    def _hash_core(self):
        return stable_hash((WeakAssignment, self.idx, self.dst, self.src))

    def __repr__(self):
        return f"WeakAssignment ({self.dst}, {self.src})"

    def __str__(self):
        return f"{self.dst!s} =W {self.src!s}"

    def replace(self, old_expr: Expression, new_expr: Expression):
        if self.dst == old_expr:
            r_dst = True
            assert isinstance(new_expr, Atom)
            replaced_dst = new_expr
        else:
            r_dst, replaced_dst = self.dst.replace(old_expr, new_expr)

        if self.src == old_expr:
            r_src = True
            replaced_src = new_expr
        else:
            r_src, replaced_src = self.src.replace(old_expr, new_expr)

        if r_dst or r_src:
            return True, WeakAssignment(self.idx, replaced_dst, replaced_src, **self.tags)
        return False, self

    def copy(self) -> WeakAssignment:
        return WeakAssignment(self.idx, self.dst, self.src, **self.tags)

    def deep_copy(self, manager) -> WeakAssignment:
        return WeakAssignment(
            manager.next_atom(), self.dst.deep_copy(manager), self.src.deep_copy(manager), **self.tags
        )


class Store(Statement):
    """
    Store statement: *addr = data
    """

    __slots__ = (
        "addr",
        "data",
        "endness",
        "guard",
        "offset",
        "size",
        "variable",
    )

    def __init__(
        self,
        idx: int | None,
        addr: Expression,
        data: Expression,
        size: int,
        endness: str,
        guard: Expression | None = None,
        variable=None,
        offset=None,
        **kwargs,
    ):
        super().__init__(idx, **kwargs)

        self.addr = addr
        self.data = data
        self.size = size
        self.endness = endness
        self.variable = variable
        self.guard = guard
        self.offset = offset  # variable_offset

    def likes(self, other):
        return (
            type(other) is Store
            and self.addr.likes(other.addr)
            and self.data.likes(other.data)
            and self.size == other.size
            and self.guard == other.guard
            and self.endness == other.endness
        )

    def matches(self, other):
        return (
            type(other) is Store
            and self.addr.matches(other.addr)
            and self.data.matches(other.data)
            and self.size == other.size
            and self.guard == other.guard
            and self.endness == other.endness
        )

    def _hash_core(self):
        return stable_hash((Store, self.idx, self.addr, self.data, self.size, self.endness, self.guard))

    def __repr__(self):
        return f"Store ({self.addr}, {self.data}[{self.size}])" + ("" if self.guard is None else f"[{self.guard}]")

    def __str__(self):
        if self.variable is None:
            return f"STORE(addr={self.addr}, data={self.data!s}, size={self.size}, endness={self.endness}, guard={self.guard})"
        return f"{self.variable.name} ={'L' if self.endness == 'Iend_LE' else 'B'} {self.data}<{self.size}>" + (
            "" if self.guard is None else f"[{self.guard}]"
        )

    def replace(self, old_expr, new_expr):
        if self.addr == old_expr:
            r_addr = True
            replaced_addr = new_expr
        else:
            r_addr, replaced_addr = self.addr.replace(old_expr, new_expr)

        if isinstance(self.data, Expression):
            if self.data == old_expr:
                r_data = True
                replaced_data = new_expr
            else:
                r_data, replaced_data = self.data.replace(old_expr, new_expr)
        else:
            r_data, replaced_data = False, self.data

        if self.guard is not None:
            r_guard, replaced_guard = self.guard.replace(old_expr, new_expr)
        else:
            r_guard, replaced_guard = False, None

        if r_addr or r_data or r_guard:
            return True, Store(
                self.idx,
                replaced_addr,
                replaced_data,
                self.size,
                self.endness,
                guard=replaced_guard,
                variable=self.variable,
                **self.tags,
            )
        return False, self

    def copy(self) -> Store:
        return Store(
            self.idx,
            self.addr,
            self.data,
            self.size,
            self.endness,
            guard=self.guard,
            variable=self.variable,
            offset=self.offset,
            **self.tags,
        )

    def deep_copy(self, manager) -> Store:
        return Store(
            manager.next_atom(),
            self.addr.deep_copy(manager),
            self.data.deep_copy(manager),
            self.size,
            self.endness,
            guard=self.guard.deep_copy(manager) if self.guard is not None else None,
            variable=self.variable,
            offset=self.offset,
            **self.tags,
        )


class Jump(Statement):
    """
    Jump statement: goto target
    """

    __slots__ = (
        "target",
        "target_idx",
    )

    def __init__(self, idx: int | None, target: Expression, target_idx: int | None = None, **kwargs):
        super().__init__(idx, **kwargs)

        self.target = target
        self.target_idx = target_idx

    def likes(self, other):
        return type(other) is Jump and is_none_or_likeable(self.target, other.target)

    def matches(self, other):
        return type(other) is Jump and is_none_or_matchable(self.target, other.target)

    def _hash_core(self):
        return stable_hash((Jump, self.idx, self.target))

    def __repr__(self):
        if self.target_idx is not None:
            return f"Jump ({self.target}.{self.target_idx})"
        return f"Jump ({self.target})"

    def __str__(self):
        if self.target_idx is not None:
            return f"Goto({self.target}.{self.target_idx})"
        return f"Goto({self.target})"

    @property
    def depth(self):
        return self.target.depth

    def replace(self, old_expr, new_expr):
        r, replaced_target = self.target.replace(old_expr, new_expr)

        if r:
            return True, Jump(self.idx, replaced_target, **self.tags)
        return False, self

    def copy(self):
        return Jump(
            self.idx,
            self.target,
            **self.tags,
        )

    def deep_copy(self, manager):
        return Jump(
            manager.next_atom(),
            self.target.deep_copy(manager),
            **self.tags,
        )


class ConditionalJump(Statement):
    """
    if (cond) {true_target} else {false_target}
    """

    __slots__ = (
        "condition",
        "false_target",
        "false_target_idx",
        "true_target",
        "true_target_idx",
    )

    def __init__(
        self,
        idx: int | None,
        condition: Expression,
        true_target: Expression | None,
        false_target: Expression | None,
        true_target_idx: int | None = None,
        false_target_idx: int | None = None,
        **kwargs,
    ):
        super().__init__(idx, **kwargs)

        self.condition = condition
        self.true_target = true_target
        self.false_target = false_target
        self.true_target_idx = true_target_idx
        self.false_target_idx = false_target_idx

    def likes(self, other):
        return (
            type(other) is ConditionalJump
            and self.condition.likes(other.condition)
            and is_none_or_likeable(self.true_target, other.true_target)
            and is_none_or_likeable(self.false_target, other.false_target)
        )

    def matches(self, other):
        return (
            type(other) is ConditionalJump
            and self.condition.matches(other.condition)
            and is_none_or_matchable(self.true_target, other.true_target)
            and is_none_or_matchable(self.false_target, other.false_target)
        )

    def _hash_core(self):
        return stable_hash(
            (
                ConditionalJump,
                self.idx,
                self.condition,
                self.true_target,
                self.false_target,
                self.true_target_idx,
                self.false_target_idx,
            )
        )

    def __repr__(self):
        return "ConditionalJump (condition: {}, true: {}{}, false: {}{})".format(
            self.condition,
            self.true_target,
            "" if self.true_target_idx is None else f".{self.true_target_idx}",
            self.false_target,
            "" if self.false_target_idx is None else f".{self.false_target_idx}",
        )

    def __str__(self):
        return "if ({}) {{ Goto {}{} }} else {{ Goto {}{} }}".format(
            self.condition,
            self.true_target,
            "" if self.true_target_idx is None else f".{self.true_target_idx}",
            self.false_target,
            "" if self.false_target_idx is None else f".{self.false_target_idx}",
        )

    def replace(self, old_expr, new_expr):
        if self.condition == old_expr:
            r_cond = True
            replaced_cond = new_expr
        else:
            r_cond, replaced_cond = self.condition.replace(old_expr, new_expr)

        if self.true_target is not None:
            if self.true_target == old_expr:
                r_true = True
                replaced_true = new_expr
            else:
                r_true, replaced_true = self.true_target.replace(old_expr, new_expr)
        else:
            r_true, replaced_true = False, self.true_target

        if self.false_target is not None:
            if self.false_target == old_expr:
                r_false = True
                replaced_false = new_expr
            else:
                r_false, replaced_false = self.false_target.replace(old_expr, new_expr)
        else:
            r_false, replaced_false = False, self.false_target

        r = r_cond or r_true or r_false

        if r:
            return True, ConditionalJump(
                self.idx,
                replaced_cond,
                replaced_true,
                replaced_false,
                true_target_idx=self.true_target_idx,
                false_target_idx=self.false_target_idx,
                **self.tags,
            )
        return False, self

    def copy(self) -> ConditionalJump:
        return ConditionalJump(
            self.idx,
            self.condition,
            self.true_target,
            self.false_target,
            true_target_idx=self.true_target_idx,
            false_target_idx=self.false_target_idx,
            **self.tags,
        )

    def deep_copy(self, manager) -> ConditionalJump:
        return ConditionalJump(
            manager.next_atom(),
            self.condition.deep_copy(manager),
            self.true_target.deep_copy(manager) if self.true_target is not None else None,
            self.false_target.deep_copy(manager) if self.false_target is not None else None,
            true_target_idx=self.true_target_idx,
            false_target_idx=self.false_target_idx,
            **self.tags,
        )


class SideEffectStatement(Statement):
    """
    A statement wrapping an expression that has side effects (e.g., a function call).

    When wrapping a Call expression, ret_expr and fp_ret_expr hold the return value destinations.
    """

    __slots__ = (
        "expr",
        "fp_ret_expr",
        "ret_expr",
    )

    def __init__(
        self,
        idx: int | None,
        expr: ailment.expression.Call,
        ret_expr: Expression | None = None,
        fp_ret_expr: Expression | None = None,
        **kwargs,
    ):
        super().__init__(idx, **kwargs)

        self.expr = expr
        self.ret_expr = ret_expr
        self.fp_ret_expr = fp_ret_expr

    @property
    def bits(self):
        return self.expr.bits if isinstance(self.expr, Expression) else 0

    @property
    def size(self):
        return self.bits // 8

    def likes(self, other):
        return (
            type(other) is SideEffectStatement
            and self.expr.likes(other.expr)
            and is_none_or_likeable(self.ret_expr, other.ret_expr)
            and is_none_or_likeable(self.fp_ret_expr, other.fp_ret_expr)
        )

    def matches(self, other):
        return (
            type(other) is SideEffectStatement
            and self.expr.matches(other.expr)
            and is_none_or_matchable(self.ret_expr, other.ret_expr)
            and is_none_or_matchable(self.fp_ret_expr, other.fp_ret_expr)
        )

    def _hash_core(self):
        return stable_hash((SideEffectStatement, self.idx, self.expr))

    def __repr__(self):
        return f"SideEffectStatement ({self.expr!r})"

    def __str__(self):
        ret_s = "no-ret-value" if self.ret_expr is None else f"{self.ret_expr}"
        fp_ret_s = "no-fp-ret-value" if self.fp_ret_expr is None else f"{self.fp_ret_expr}"
        return f"{self.expr!s} ret: {ret_s}, fp_ret: {fp_ret_s}"

    def replace(self, old_expr: Expression, new_expr: Expression):
        if self.expr == old_expr:
            r_expr = True
            replaced_expr = new_expr
        else:
            r_expr, replaced_expr = self.expr.replace(old_expr, new_expr)

        r = r_expr

        new_ret_expr = self.ret_expr
        if self.ret_expr:
            if self.ret_expr == old_expr:
                r_ret = True
                replaced_ret = new_expr
            else:
                r_ret, replaced_ret = self.ret_expr.replace(old_expr, new_expr)
            r |= r_ret
            new_ret_expr = replaced_ret

        new_fp_ret_expr = self.fp_ret_expr
        if self.fp_ret_expr:
            if self.fp_ret_expr == old_expr:
                r_ret = True
                replaced_fp_ret = new_expr
            else:
                r_ret, replaced_fp_ret = self.fp_ret_expr.replace(old_expr, new_expr)
            r |= r_ret
            new_fp_ret_expr = replaced_fp_ret

        if r:
            return True, SideEffectStatement(
                self.idx,
                replaced_expr,
                ret_expr=new_ret_expr,
                fp_ret_expr=new_fp_ret_expr,
                **self.tags,
            )
        return False, self

    def copy(self) -> SideEffectStatement:
        return SideEffectStatement(
            self.idx,
            self.expr.copy(),
            ret_expr=self.ret_expr,
            fp_ret_expr=self.fp_ret_expr,
            **self.tags,
        )

    def deep_copy(self, manager) -> SideEffectStatement:
        return SideEffectStatement(
            manager.next_atom(),
            self.expr.deep_copy(manager),
            ret_expr=self.ret_expr.deep_copy(manager) if self.ret_expr is not None else None,
            fp_ret_expr=self.fp_ret_expr.deep_copy(manager) if self.fp_ret_expr is not None else None,
            **self.tags,
        )


class Return(Statement):
    """
    Return statement: (return expr_a), (return)
    """

    __slots__ = ("ret_exprs",)

    def __init__(self, idx: int | None, ret_exprs: Iterable[Expression], **kwargs):
        super().__init__(idx, **kwargs)
        self.ret_exprs = ret_exprs if isinstance(ret_exprs, list) else list(ret_exprs)

    def likes(self, other):
        return type(other) is Return and is_none_or_likeable(self.ret_exprs, other.ret_exprs, is_list=True)

    def matches(self, other):
        return type(other) is Return and is_none_or_matchable(self.ret_exprs, other.ret_exprs, is_list=True)

    def _hash_core(self):
        return stable_hash((Return, self.idx, tuple(self.ret_exprs)))

    def __repr__(self):
        return "Return to ({})".format(",".join(repr(x) for x in self.ret_exprs))

    def __str__(self):
        exprs = ",".join(str(ret_expr) for ret_expr in self.ret_exprs)
        if not exprs:
            return "return;"
        return f"return {exprs};"

    def replace(self, old_expr, new_expr):
        new_ret_exprs = []
        replaced = False

        for expr in self.ret_exprs:
            if expr == old_expr:
                r_expr = True
                replaced_expr = new_expr
            else:
                r_expr, replaced_expr = expr.replace(old_expr, new_expr)
            if r_expr:
                replaced = True
                new_ret_exprs.append(replaced_expr)
            else:
                new_ret_exprs.append(expr)

        if replaced:
            return True, Return(
                self.idx,
                new_ret_exprs,
                **self.tags,
            )

        return False, self

    def copy(self):
        return Return(
            self.idx,
            self.ret_exprs[::],
            **self.tags,
        )

    def deep_copy(self, manager):
        return Return(
            manager.next_atom(),
            [expr.deep_copy(manager) for expr in self.ret_exprs],
            **self.tags,
        )


class CAS(Statement):
    """
    Atomic compare-and-swap.

    *_lo and *_hi are used to represent the low and high parts of a 128-bit CAS operation; *_hi is None if the CAS
    operation works on values that are less than or equal to 64 bits.

    addr: The address to be compared and swapped.
    data: The value to be written if the comparison is successful.
    expd: The expected value to be compared against.
    old: The value that is currently stored at addr before compare-and-swap; it will be returned after compare-and-swap.
    """

    __slots__ = ("addr", "data_hi", "data_lo", "endness", "expd_hi", "expd_lo", "old_hi", "old_lo")

    def __init__(
        self,
        idx: int | None,
        addr: Expression,
        data_lo: Expression,
        data_hi: Expression | None,
        expd_lo: Expression,
        expd_hi: Expression | None,
        old_lo: Atom,
        old_hi: Atom | None,
        endness: str,
        **kwargs,
    ):
        super().__init__(idx, **kwargs)
        self.addr = addr
        self.data_lo = data_lo
        self.data_hi = data_hi
        self.expd_lo = expd_lo
        self.expd_hi = expd_hi
        self.old_lo = old_lo
        self.old_hi = old_hi
        self.endness = endness

    def _hash_core(self):
        return stable_hash(
            (
                CAS,
                self.idx,
                self.addr,
                self.data_lo,
                self.data_hi,
                self.expd_lo,
                self.expd_hi,
                self.old_lo,
                self.old_hi,
                self.endness,
            )
        )

    def __repr__(self):
        if self.old_hi is None:
            return f"CAS({self.addr}, {self.data_lo}, {self.expd_lo}, {self.old_lo})"
        return (
            f"CAS({self.addr}, {self.data_hi} .. {self.data_lo}, {self.expd_hi} .. {self.expd_lo}, "
            f"{self.old_hi} .. {self.old_lo})"
        )

    def __str__(self):
        if self.old_hi is None:
            return f"{self.old_lo} = CAS({self.addr}, {self.data_lo}, {self.expd_lo})"
        return (
            f"{self.old_hi} .. {self.old_lo} = CAS({self.addr}, {self.data_hi} .. {self.data_lo}, "
            f"{self.expd_hi} .. {self.expd_lo})"
        )

    def replace(self, old_expr: Expression, new_expr: Expression) -> tuple[bool, CAS]:
        r_addr, replaced_addr = self.addr.replace(old_expr, new_expr)
        r_data_lo, replaced_data_lo = self.data_lo.replace(old_expr, new_expr)
        r_data_hi, replaced_data_hi = self.data_hi.replace(old_expr, new_expr) if self.data_hi else (False, None)
        r_expd_lo, replaced_expd_lo = self.expd_lo.replace(old_expr, new_expr)
        r_expd_hi, replaced_expd_hi = self.expd_hi.replace(old_expr, new_expr) if self.expd_hi else (False, None)
        r_old_lo, replaced_old_lo = self.old_lo.replace(old_expr, new_expr)
        r_old_hi, replaced_old_hi = self.old_hi.replace(old_expr, new_expr) if self.old_hi else (False, None)

        if r_addr or r_data_lo or r_data_hi or r_expd_lo or r_expd_hi or r_old_lo or r_old_hi:
            return True, CAS(
                self.idx,
                replaced_addr,
                replaced_data_lo,
                replaced_data_hi,
                replaced_expd_lo,
                replaced_expd_hi,
                replaced_old_lo,
                replaced_old_hi,
                endness=self.endness,
                **self.tags,
            )
        return False, self

    def copy(self) -> CAS:
        return CAS(
            self.idx,
            self.addr,
            self.data_lo,
            self.data_hi,
            self.expd_lo,
            self.expd_hi,
            self.old_lo,
            self.old_hi,
            endness=self.endness,
            **self.tags,
        )

    def deep_copy(self, manager) -> CAS:
        return CAS(
            manager.next_atom(),
            self.addr.deep_copy(manager),
            self.data_lo.deep_copy(manager),
            self.data_hi.deep_copy(manager) if self.data_hi is not None else None,
            self.expd_lo.deep_copy(manager),
            self.expd_hi.deep_copy(manager) if self.expd_hi is not None else None,
            self.old_lo.deep_copy(manager),
            self.old_hi.deep_copy(manager) if self.old_hi is not None else None,
            endness=self.endness,
            **self.tags,
        )

    def likes(self, other) -> bool:
        return (
            type(other) is CAS
            and self.addr.likes(other.addr)
            and self.data_lo.likes(other.data_lo)
            and (self.data_hi is None or self.data_hi.likes(other.data_hi))
            and self.expd_lo.likes(other.expd_lo)
            and (self.expd_hi is None or self.expd_hi.likes(other.expd_hi))
            and self.old_lo.likes(other.old_lo)
            and (self.old_hi is None or self.old_hi.likes(other.old_hi))
        )

    def matches(self, other) -> bool:
        return (
            type(other) is CAS
            and self.addr.matches(other.addr)
            and self.data_lo.matches(other.data_lo)
            and (self.data_hi is None or self.data_hi.matches(other.data_hi))
            and self.expd_lo.matches(other.expd_lo)
            and (self.expd_hi is None or self.expd_hi.matches(other.expd_hi))
            and self.old_lo.matches(other.old_lo)
            and (self.old_hi is None or self.old_hi.matches(other.old_hi))
        )

    @property
    def bits(self) -> int:
        if self.old_hi is None:
            return self.old_lo.bits
        return self.old_lo.bits + self.old_hi.bits

    @property
    def size(self) -> int:
        return self.bits // 8


class DirtyStatement(Statement):
    """
    Wrapper around the original statement, which is usually not convertible (temporarily).
    """

    __slots__ = ("dirty",)

    def __init__(self, idx: int | None, dirty: DirtyExpression, **kwargs):
        super().__init__(idx, **kwargs)
        self.dirty = dirty

    def _hash_core(self):
        return stable_hash((DirtyStatement, self.dirty))

    def __repr__(self):
        return repr(self.dirty)

    def __str__(self):
        return str(self.dirty)

    def replace(self, old_expr, new_expr):
        if self.dirty == old_expr:
            assert isinstance(new_expr, DirtyExpression)
            return True, DirtyStatement(self.idx, new_expr, **self.tags)
        r, new_dirty = self.dirty.replace(old_expr, new_expr)
        if r:
            return True, DirtyStatement(self.idx, new_dirty, **self.tags)
        return False, self

    def copy(self) -> DirtyStatement:
        return DirtyStatement(self.idx, self.dirty, **self.tags)

    def deep_copy(self, manager) -> DirtyStatement:
        return DirtyStatement(manager.next_atom(), self.dirty.deep_copy(manager), **self.tags)

    def likes(self, other):
        return type(other) is DirtyStatement and self.dirty.likes(other.dirty)

    def matches(self, other):
        return type(other) is DirtyStatement and self.dirty.matches(other.dirty)


class Label(Statement):
    """
    A dummy statement that indicates a label with a name.
    """

    __slots__ = ("name",)

    def __init__(self, idx: int | None, name: str, **kwargs):
        super().__init__(idx, **kwargs)
        self.name = name

    def likes(self, other: Label):
        return isinstance(other, Label)

    def replace(self, old_expr, new_expr):
        return False, self

    matches = likes

    def _hash_core(self):
        return stable_hash(
            (
                Label,
                self.name,
            )
        )

    def __repr__(self):
        return f"Label {self.name}"

    def __str__(self):
        return f"{self.name}:"

    def copy(self) -> Label:
        return Label(self.idx, self.name, **self.tags)

    def deep_copy(self, manager) -> Label:
        return Label(manager.next_atom(), self.name, **self.tags)



from __future__ import annotations
import claripy

import angr


class iscntrl(angr.SimProcedure):
    # pylint: disable=arguments-differ, missing-class-docstring
    def run(self, c):
        is_low_ctrl = claripy.And(c >= 0, c <= 31)
        is_del = c == 127
        int_size = self.arch.sizeof["int"]  # type: ignore[reportOptionalMemberAccess]

        return claripy.If(
            claripy.Or(is_low_ctrl, is_del),
            claripy.BVV(1, int_size),
            claripy.BVV(0, int_size),
        )



from __future__ import annotations

import json
import logging
import os
import re

try:
    import litellm  # type: ignore
except ImportError:
    litellm = None

l = logging.getLogger(name=__name__)


class LLMClient:
    """
    A client for interacting with LLMs via LiteLLM.
    Used by the decompiler to suggest improved variable names, function names, and types.
    """

    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        api_base: str | None = None,
        max_tokens: int = 4096,
        temperature: float = 0.0,
    ):
        if litellm is None:
            raise ImportError(
                "litellm is required for LLM support. Install it with: pip install angr[llm]  or  pip install litellm"
            )

        self.model = model
        self.api_key = api_key
        self.api_base = api_base
        self.max_tokens = max_tokens
        self.temperature = temperature

    @classmethod
    def from_env(cls) -> LLMClient | None:
        """
        Create an LLMClient from environment variables.

        Uses ANGR_LLM_MODEL (required), ANGR_LLM_API_KEY (optional), and ANGR_LLM_API_BASE (optional).
        Returns None if ANGR_LLM_MODEL is not set.
        """
        model = os.environ.get("ANGR_LLM_MODEL")
        if not model:
            return None
        api_key = os.environ.get("ANGR_LLM_API_KEY")
        api_base = os.environ.get("ANGR_LLM_API_BASE")
        return cls(model=model, api_key=api_key, api_base=api_base)

    def completion(self, messages: list[dict[str, str]], **kwargs) -> str:
        """
        Call the LLM with the given messages and return the response text.
        """
        call_kwargs: dict = {
            "model": self.model,
            "messages": messages,
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
        }
        if self.api_key:
            call_kwargs["api_key"] = self.api_key
        if self.api_base:
            call_kwargs["api_base"] = self.api_base
        call_kwargs.update(kwargs)

        response = litellm.completion(**call_kwargs)  # type: ignore
        return response.choices[0].message.content

    def completion_json(self, messages: list[dict[str, str]], **kwargs) -> dict | None:
        """
        Call the LLM and parse the response as JSON.
        Strips markdown code fences if present. Returns None on parse failure.
        """
        text = self.completion(messages, **kwargs)
        if not text:
            return None

        # strip markdown code fences
        text = text.strip()
        text = re.sub(r"^```(?:json)?\s*\n?", "", text)
        text = re.sub(r"\n?```\s*$", "", text)
        text = text.strip()

        try:
            return json.loads(text)
        except (json.JSONDecodeError, ValueError):
            l.warning("Failed to parse LLM response as JSON: %s", text[:200])
            return None

    def __repr__(self):
        return f"<LLMClient model={self.model!r}>"



from __future__ import annotations

from .base import SimConcretizationStrategy


class SimConcretizationStrategyNonzero(SimConcretizationStrategy):
    """
    Concretization strategy that returns any non-zero solution.
    """

    def _concretize(self, memory, addr, **kwargs):
        child_constraints = (addr != 0,)
        extra_constraints = kwargs.pop("extra_constraints", None)
        if extra_constraints is not None:
            child_constraints += tuple(extra_constraints)
        return [self._any(memory, addr, extra_constraints=child_constraints, **kwargs)]



# pylint:disable=wrong-import-position,broad-exception-caught,ungrouped-imports,import-outside-toplevel
from __future__ import annotations
import pathlib
import copy
from types import FunctionType
from typing import Any, TYPE_CHECKING
from collections.abc import Iterable
import logging

import networkx

import angr
from angr import ailment
from angr.ailment.block import Block
from angr.analyses.decompiler.counters.call_counter import AILBlockCallCounter
from angr.analyses.decompiler.peephole_optimizations.base import (
    PeepholeOptimizationExprBase,
    PeepholeOptimizationMultiStmtBase,
)
from angr.utils.ail import is_phi_assignment
from .seq_to_blocks import SequenceToBlocks

if TYPE_CHECKING:
    from angr.ailment import Address

_l = logging.getLogger(__name__)


def remove_last_statement(node):
    stmt = None

    if type(node) is CodeNode:
        stmt = remove_last_statement(node.node)
    elif type(node) is ailment.Block:
        stmt = node.statements[-1]
        node.statements = node.statements[:-1]
    elif type(node) is MultiNode or type(node) is SequenceNode:
        if node.nodes:
            stmt = remove_last_statement(node.nodes[-1])
            if BaseNode.test_empty_node(node.nodes[-1]):
                node.nodes = node.nodes[:-1]
    elif type(node) is ConditionNode:
        if node.true_node is None and node.false_node is not None:
            stmt = remove_last_statement(node.false_node)
        elif node.true_node is not None and node.false_node is None:
            stmt = remove_last_statement(node.true_node)
        else:
            raise NotImplementedError("More than one last statement exist")
    elif type(node) is LoopNode:
        stmt = remove_last_statement(node.sequence_node)
    else:
        raise NotImplementedError(type(node))

    return stmt


def remove_last_statements(node) -> bool:
    if type(node) is CodeNode:
        return remove_last_statements(node.node)
    if type(node) is ailment.Block:
        if not node.statements:
            return False
        node.statements = node.statements[:-1]
        return True
    if type(node) is MultiNode or type(node) is SequenceNode:
        if node.nodes:
            remove_last_statements(node.nodes[-1])
            if BaseNode.test_empty_node(node.nodes[-1]):
                node.nodes = node.nodes[:-1]
            return True
        return False
    if type(node) is ConditionNode:
        r = False
        if node.true_node is None and node.false_node is not None:
            r |= remove_last_statements(node.false_node)
        if node.true_node is not None and node.false_node is None:
            r |= remove_last_statements(node.true_node)
        return r
    if type(node) is LoopNode:
        return remove_last_statements(node.sequence_node)
    raise NotImplementedError(type(node))


def append_statement(node, stmt):
    if type(node) is CodeNode:
        append_statement(node.node, stmt)
        return
    if type(node) is ailment.Block:
        node.statements.append(stmt)
        return
    if type(node) is MultiNode:
        if node.nodes:
            append_statement(node.nodes[-1], stmt)
        else:
            raise NotImplementedError("MultiNode without nodes")
        return
    if type(node) is SequenceNode:
        if node.nodes:
            append_statement(node.nodes[-1], stmt)
        else:
            raise NotImplementedError("SequenceNode without nodes")
        return

    raise NotImplementedError(type(node))


def replace_last_statement(node, old_stmt, new_stmt):
    if type(node) is CodeNode:
        replace_last_statement(node.node, old_stmt, new_stmt)
        return
    if type(node) is ailment.Block:
        if node.statements[-1] is old_stmt:
            node.statements[-1] = new_stmt
        return
    if type(node) is MultiNode:
        if node.nodes:
            replace_last_statement(node.nodes[-1], old_stmt, new_stmt)
        return
    if type(node) is SequenceNode:
        if node.nodes:
            replace_last_statement(node.nodes[-1], old_stmt, new_stmt)
        return
    if type(node) is ConditionNode:
        if node.true_node is not None:
            replace_last_statement(node.true_node, old_stmt, new_stmt)
        if node.false_node is not None:
            replace_last_statement(node.false_node, old_stmt, new_stmt)
        return

    raise NotImplementedError(type(node))


def extract_jump_targets(stmt):
    """
    Extract concrete goto targets from a Jump or a ConditionalJump statement.

    :param stmt:    The statement to analyze.
    :return:        A list of known concrete jump targets.
    :rtype:         list
    """

    targets = []

    if isinstance(stmt, ailment.Stmt.Jump):
        if isinstance(stmt.target, ailment.Expr.Const):
            targets.append(stmt.target.value)
    elif isinstance(stmt, ailment.Stmt.ConditionalJump):
        if isinstance(stmt.true_target, ailment.Expr.Const):
            targets.append(stmt.true_target.value)
        if isinstance(stmt.false_target, ailment.Expr.Const):
            targets.append(stmt.false_target.value)

    return targets


def switch_extract_cmp_bounds(
    last_stmt: ailment.Stmt.ConditionalJump | ailment.Stmt.Statement,
) -> tuple[Any, int, int] | None:
    """
    Check the last statement of the switch-case header node, and extract lower+upper bounds for the comparison.

    :param last_stmt:   The last statement of the switch-case header node.
    :return:            A tuple of (comparison expression, lower bound, upper bound), or None
    """

    if not isinstance(last_stmt, ailment.Stmt.ConditionalJump):
        return None
    return switch_extract_cmp_bounds_from_condition(last_stmt.condition)


def switch_extract_cmp_bounds_from_condition(cond: ailment.Expr.Expression) -> tuple[Any, int, int] | None:
    # TODO: Add more operations
    if isinstance(cond, ailment.Expr.BinaryOp):
        op = cond.op
        op0, op1 = cond.operands
        if not isinstance(op1, ailment.Expr.Const):
            # swap them
            match op:
                case "CmpLE":
                    op = "CmpGE"
                case "CmpLT":
                    op = "CmpGT"
                case "CmpGE":
                    op = "CmpLE"
                case "CmpGT":
                    op = "CmpLT"
                case _:
                    # unsupported
                    return None
            op0, op1 = op1, op0

        if op in {"CmpLE", "CmpLT"}:
            if not (isinstance(op1, ailment.Expr.Const) and isinstance(op1.value, int)):
                return None
            cmp_ub = op1.value if op == "CmpLE" else op1.value - 1
            cmp_lb = 0
            cmp = op0
            if (
                isinstance(cmp, ailment.Expr.BinaryOp)
                and cmp.op == "Sub"
                and isinstance(cmp.operands[1], ailment.Expr.Const)
                and isinstance(cmp.operands[1].value, int)
            ):
                cmp_ub += cmp.operands[1].value
                cmp_lb += cmp.operands[1].value
                cmp = cmp.operands[0]
            return cmp, cmp_lb, cmp_ub

        if op in {"CmpGE", "CmpGT"}:
            # We got the negated condition here
            #  CmpGE -> CmpLT
            #  CmpGT -> CmpLE
            if not (isinstance(op1, ailment.Expr.Const) and isinstance(op1.value, int)):
                return None
            cmp_ub = op1.value if op == "CmpGT" else op1.value - 1
            cmp_lb = 0
            cmp = op0
            if (
                isinstance(cmp, ailment.Expr.BinaryOp)
                and cmp.op == "Sub"
                and isinstance(cmp.operands[1], ailment.Expr.Const)
                and isinstance(cmp.operands[1].value, int)
            ):
                cmp_ub += cmp.operands[1].value
                cmp_lb += cmp.operands[1].value
                cmp = cmp.operands[0]
            return cmp, cmp_lb, cmp_ub

    return None


def switch_extract_switch_expr_from_jump_target(target: ailment.Expr.Expression) -> ailment.Expr.Expression | None:
    """
    Extract the switch expression from the indirect jump target expression.

    :param target:  The target of the indirect jump statement.
    :return:        The extracted expression if successful, or None otherwise.
    """

    # e.g.: Jump (Conv(32->64, (Load(addr=((0x140000000<64> + (vvar_229{reg 80} * 0x4<64>)) + 0x2290<64>),
    #               size=4,
    #               endness=Iend_LE
    #             ) + 0x140000000<32>)))

    found_load = False
    while True:
        if isinstance(target, ailment.Expr.Convert):
            if target.from_bits < target.to_bits:
                target = target.operand
            else:
                return None
        elif isinstance(target, ailment.Expr.BinaryOp):
            if target.op == "Add":
                # it must be adding the target expr with a constant
                if isinstance(target.operands[0], ailment.Expr.Const):
                    target = target.operands[1]
                elif isinstance(target.operands[1], ailment.Expr.Const):
                    target = target.operands[0]
                else:
                    return None
            elif target.op == "Mul":
                # it must be multiplying the target expr with a constant
                if isinstance(target.operands[0], ailment.Expr.Const):
                    target = target.operands[1]
                elif isinstance(target.operands[1], ailment.Expr.Const):
                    target = target.operands[0]
                else:
                    return None
            elif target.op == "And":
                # it must be and-ing the target expr with a constant
                if (
                    isinstance(target.operands[1], ailment.Expr.VirtualVariable)
                    and isinstance(target.operands[0], ailment.Expr.Const)
                ) or (
                    isinstance(target.operands[0], ailment.Expr.VirtualVariable)
                    and isinstance(target.operands[1], ailment.Expr.Const)
                ):
                    break
                return None
            else:
                return None
        elif isinstance(target, ailment.Expr.Load):
            # we want the address!
            found_load = True
            target = target.addr
        elif isinstance(target, ailment.Expr.VirtualVariable):
            break
        else:
            return None
    return target if found_load else None


def switch_extract_bitwiseand_jumptable_info(last_stmt: ailment.Stmt.Jump) -> tuple[Any, int, int] | None:
    """
    Check the last statement of the switch-case header node (whose address is loaded from a jump table and computed
    using an index) and extract necessary information for rebuilding the switch-case construct.

    An example of the statement:

    Goto(Conv(32->s64, (
        Load(addr=(0x4530e4<64> + (Conv(32->64, (Conv(64->32, vvar_287{reg 32}) & 0x3<32>)) * 0x4<64>)),
             size=4, endness=Iend_LE) + 0x4530e4<32>))
    )

    Another example:

    Load(addr=(((vvar_9{reg 36} & 0x3<32>) * 0x4<32>) + 0x42cd28<32>), size=4, endness=Iend_LE)

    :param last_stmt:   The last statement of the switch-case header node.
    :return:            A tuple of (index expression, lower bound, upper bound), or None
    """

    if not isinstance(last_stmt, ailment.Stmt.Jump):
        return None

    # unpack the target expression
    target = last_stmt.target
    jump_addr_offset = None
    jumptable_load_addr = None
    while True:
        if isinstance(target, ailment.Expr.Convert) and (
            (target.from_bits == 32 and target.to_bits == 64) or (target.from_bits == 16 and target.to_bits == 32)
        ):
            target = target.operand
            continue
        if isinstance(target, ailment.Expr.BinaryOp) and target.op == "Add":
            if isinstance(target.operands[0], ailment.Expr.Const) and isinstance(target.operands[1], ailment.Expr.Load):
                jump_addr_offset = target.operands[0].value
                jumptable_load_addr = target.operands[1].addr
                break
            if isinstance(target.operands[1], ailment.Expr.Const) and isinstance(target.operands[0], ailment.Expr.Load):
                jump_addr_offset = target.operands[1].value
                jumptable_load_addr = target.operands[0].addr
                break
            return None
        if isinstance(target, ailment.Expr.Const):
            return None
        if isinstance(target, ailment.Expr.Load):
            jumptable_load_addr = target.addr
            jump_addr_offset = 0
            break
        break

    if jump_addr_offset is None or jumptable_load_addr is None:
        return None

    # parse jumptable_load_addr
    jumptable_offset = None
    jumptable_base_addr = None
    if isinstance(jumptable_load_addr, ailment.Expr.BinaryOp) and jumptable_load_addr.op == "Add":
        if isinstance(jumptable_load_addr.operands[0], ailment.Expr.Const):
            jumptable_base_addr = jumptable_load_addr.operands[0]
            jumptable_offset = jumptable_load_addr.operands[1]
        elif isinstance(jumptable_load_addr.operands[1], ailment.Expr.Const):
            jumptable_offset = jumptable_load_addr.operands[0]
            jumptable_base_addr = jumptable_load_addr.operands[1]

    if jumptable_offset is None or jumptable_base_addr is None:
        return None

    # parse jumptable_offset
    expr = jumptable_offset
    coeff = None
    index_expr = None
    lb = None
    ub: int | None = None
    while expr is not None:
        if isinstance(expr, ailment.Expr.BinaryOp):
            if expr.op == "Mul":
                if isinstance(expr.operands[1], ailment.Expr.Const):
                    coeff = expr.operands[1].value
                    expr = expr.operands[0]
                elif isinstance(expr.operands[0], ailment.Expr.Const):
                    coeff = expr.operands[0].value
                    expr = expr.operands[1]
                else:
                    return None
            elif expr.op == "And":
                masks = {0x1, 0x3, 0x7, 0xF, 0x1F, 0x3F, 0x7F, 0xFF, 0x1FF, 0x3FF}
                if isinstance(expr.operands[1], ailment.Expr.Const) and expr.operands[1].value in masks:
                    lb = 0
                    ub = expr.operands[1].value  # type: ignore
                    index_expr = expr
                    break
                if isinstance(expr.operands[0], ailment.Expr.Const) and expr.operands[1].value in masks:
                    lb = 0
                    ub = expr.operands[0].value  # type: ignore
                    index_expr = expr
                    break
                return None
            else:
                return None
        elif isinstance(expr, ailment.Expr.Convert):
            if expr.is_signed is False:
                expr = expr.operand
            else:
                return None
        else:
            break

    if coeff is not None and index_expr is not None and lb is not None and ub is not None:
        return index_expr, lb, ub
    return None


def get_ast_subexprs(claripy_ast):
    queue = [claripy_ast]
    while queue:
        ast = queue.pop(0)
        if ast.op == "And":
            queue += ast.args[1:]
            yield ast.args[0]
        else:
            yield ast


def insert_node(parent, insert_location: str, node, node_idx: int, label=None):
    if insert_location not in {"before", "after"}:
        raise ValueError('"insert_location" must be either "before" or "after"')

    if isinstance(parent, SequenceNode):
        if insert_location == "before":
            parent.nodes.insert(node_idx, node)
        else:  # if insert_location == "after":
            parent.nodes.insert(node_idx + 1, node)
    elif isinstance(parent, CodeNode):
        # Make a new sequence node
        if insert_location == "before":
            seq = SequenceNode(parent.addr, nodes=[node, parent.node])
        else:  # if insert_location == "after":
            seq = SequenceNode(parent.addr, nodes=[parent.node, node])
        parent.node = seq
    elif isinstance(parent, MultiNode):
        if insert_location == "before":
            parent.nodes.insert(node_idx, node)
        else:
            parent.nodes.insert(node_idx + 1, node)
    elif isinstance(parent, ConditionNode):
        if node_idx == 0:
            # true node
            if not isinstance(parent.true_node, SequenceNode):
                if parent.true_node is None:
                    parent.true_node = SequenceNode(parent.addr, nodes=[])
                else:
                    parent.true_node = SequenceNode(parent.true_node.addr, nodes=[parent.true_node])
            insert_node(parent.true_node, insert_location, node, 0)
        else:
            # false node
            if not isinstance(parent.false_node, SequenceNode):
                if parent.false_node is None:
                    parent.false_node = SequenceNode(parent.addr, nodes=[])
                else:
                    parent.false_node = SequenceNode(parent.false_node.addr, nodes=[parent.false_node])
            insert_node(parent.false_node, insert_location, node, 0)
    elif isinstance(parent, CascadingConditionNode):
        cond, child_node = parent.condition_and_nodes[node_idx]
        if not isinstance(child_node, SequenceNode):
            child_node = SequenceNode(child_node.addr, nodes=[child_node])
            parent.condition_and_nodes[node_idx] = (cond, child_node)
        insert_node(child_node, insert_location, node, 0)
    elif isinstance(parent, SwitchCaseNode):
        # note that this case will be hit only when the parent node is not a container, such as SequenceNode or
        # MultiNode. we always need to create a new SequenceNode and replace the original node in place.

        if label == "switch_expr":
            raise TypeError("You cannot insert a node after an expression.")
        if label == "case":
            # node_idx is the case number.
            if insert_location == "after":
                new_nodes = [parent.cases[node_idx], node]
            elif insert_location == "before":
                new_nodes = [node, parent.cases[node_idx]]
            else:
                raise TypeError(f'Unsupported insert_location value "{insert_location}".')
            seq = SequenceNode(new_nodes[0].addr, nodes=new_nodes)
            parent.cases[node_idx] = seq
        elif label == "default":
            if insert_location == "after":
                new_nodes = [parent.default_node, node]
            elif insert_location == "before":
                new_nodes = [node, parent.default_node]
            else:
                raise TypeError(f"Unsupported 'insert_location' value {insert_location!r}.")
            seq = SequenceNode(new_nodes[0].addr, nodes=new_nodes)
            parent.default_node = seq
        else:
            raise TypeError(
                f'Unsupported label value "{label}". Must be one of the following: switch_expr, case, default.'
            )
    elif isinstance(parent, LoopNode):
        if label == "condition":
            raise ValueError("Cannot insert nodes into a condition expression.")
        if label == "body":
            if not isinstance(parent.sequence_node, SequenceNode):
                parent.sequence_node = SequenceNode(parent.sequence_node.addr, nodes=[parent.sequence_node])
            insert_node(parent.sequence_node, insert_location, node, node_idx)
        else:
            raise NotImplementedError(label)
    else:
        raise NotImplementedError(type(parent))


def _merge_ail_nodes(graph, node_a: ailment.Block, node_b: ailment.Block) -> ailment.Block:
    in_edges = list(graph.in_edges(node_a, data=True))
    out_edges = list(graph.out_edges(node_b, data=True))

    a_ogs = graph.nodes[node_a].get("original_nodes", [])
    b_ogs = graph.nodes[node_b].get("original_nodes", [])
    new_node = node_a.copy() if node_a.addr <= node_b.addr else node_b.copy()
    old_node = node_b if new_node == node_a else node_a
    # remove jumps in the middle of nodes when merging
    if new_node.statements and isinstance(new_node.statements[-1], ailment.Stmt.Jump):
        new_node.statements = new_node.statements[:-1]
    new_node.statements += old_node.statements
    new_node.original_size += old_node.original_size

    graph.remove_node(node_a)
    graph.remove_node(node_b)

    if new_node is not None:
        graph.add_node(new_node, original_nodes=a_ogs + b_ogs)
        for src, _, data in in_edges:
            if src is node_b:
                src = new_node
            graph.add_edge(src, new_node, **data)

        for _, dst, data in out_edges:
            if dst is node_a:
                dst = new_node
            graph.add_edge(new_node, dst, **data)

    return new_node


def to_ail_supergraph(transition_graph: networkx.DiGraph, allow_fake=False) -> networkx.DiGraph:
    """
    Takes an AIL graph and converts it into a AIL graph that treats calls and redundant jumps
    as parts of a bigger block instead of transitions. Calls to returning functions do not terminate basic blocks.

    Based on region_identifier super_graph

    :return: A converted super transition graph
    """
    # make a copy of the graph
    transition_graph = networkx.DiGraph(transition_graph)
    networkx.set_node_attributes(transition_graph, {node: [node] for node in transition_graph.nodes}, "original_nodes")

    while True:
        for src, dst, data in transition_graph.edges(data=True):
            type_ = data.get("type", None)

            if len(list(transition_graph.successors(src))) == 1 and len(list(transition_graph.predecessors(dst))) == 1:
                # calls in the middle of blocks OR boring jumps
                if (type_ == "fake_return") or (src.addr + src.original_size == dst.addr) or allow_fake:
                    _merge_ail_nodes(transition_graph, src, dst)
                    break

            # calls to functions with no return
            elif type_ == "call":
                transition_graph.remove_node(dst)
                break
        else:
            break

    return transition_graph


def is_empty_node(node) -> bool:
    if isinstance(node, ailment.Block):
        return not node.statements
    if isinstance(node, MultiNode):
        return all(is_empty_node(n) for n in node.nodes)
    if isinstance(node, SequenceNode):
        return all(is_empty_node(n) for n in node.nodes)
    return False


def is_empty_or_label_only_node(node) -> bool:
    if isinstance(node, ailment.Block):
        return not has_nonlabel_statements(node)
    if isinstance(node, MultiNode):
        return all(is_empty_or_label_only_node(n) for n in node.nodes)
    if isinstance(node, SequenceNode):
        return all(is_empty_or_label_only_node(n) for n in node.nodes)
    return False


def has_nonlabel_statements(block: ailment.Block) -> bool:
    return bool(block.statements and any(not isinstance(stmt, ailment.Stmt.Label) for stmt in block.statements))


def has_nonlabel_nonphi_statements(block: ailment.Block) -> bool:
    return bool(
        block.statements
        and any(not (isinstance(stmt, ailment.Stmt.Label) or is_phi_assignment(stmt)) for stmt in block.statements)
    )


def first_nonlabel_statement(block: ailment.Block | MultiNode) -> ailment.Stmt.Statement | None:
    if isinstance(block, MultiNode):
        for n in block.nodes:
            stmt = first_nonlabel_statement(n)
            if stmt is not None:
                return stmt
        return None

    for stmt in block.statements:
        if not isinstance(stmt, ailment.Stmt.Label):
            return stmt
    return None


def first_nonlabel_statement_id(block: ailment.Block) -> int | None:
    for idx, stmt in enumerate(block.statements):
        if not isinstance(stmt, ailment.Stmt.Label):
            return idx
    return len(block.statements)


def first_nonlabel_nonphi_statement(block: ailment.Block | MultiNode) -> ailment.Stmt.Statement | None:
    if isinstance(block, MultiNode):
        for n in block.nodes:
            stmt = first_nonlabel_nonphi_statement(n)
            if stmt is not None:
                return stmt
        return None

    for stmt in block.statements:
        if not (isinstance(stmt, ailment.Stmt.Label) or is_phi_assignment(stmt)):
            return stmt
    return None


def last_nonlabel_statement(block: ailment.Block) -> ailment.Stmt.Statement | None:
    for stmt in reversed(block.statements):
        if not isinstance(stmt, ailment.Stmt.Label):
            return stmt
    return None


def last_node(node: BaseNode) -> BaseNode | ailment.Block | None:
    """
    Get the last node in a sequence or code node.
    """
    if isinstance(node, CodeNode):
        return last_node(node.node)
    if isinstance(node, SequenceNode):
        if not node.nodes:
            return None
        return last_node(node.nodes[-1])
    return node


def first_nonlabel_node(seq: SequenceNode) -> BaseNode | ailment.Block | None:
    for node in seq.nodes:
        inner_node = node.node if isinstance(node, CodeNode) else node
        if isinstance(inner_node, ailment.Block) and not has_nonlabel_statements(inner_node):
            continue
        return node
    return None


def first_nonlabel_nonphi_node(seq: SequenceNode) -> BaseNode | ailment.Block | None:
    for node in seq.nodes:
        inner_node = node.node if isinstance(node, CodeNode) else node
        if isinstance(inner_node, ailment.Block) and not has_nonlabel_nonphi_statements(inner_node):
            continue
        return node
    return None


def remove_labels(graph: networkx.DiGraph):
    new_graph = networkx.DiGraph()
    nodes_map = {}
    for node in graph:
        node_copy = node.copy()
        node_copy.statements = [stmt for stmt in node_copy.statements if not isinstance(stmt, ailment.Stmt.Label)]
        nodes_map[node] = node_copy

    for old_node in graph.nodes:
        new_graph.add_node(nodes_map[old_node])

    for src, dst, data in graph.edges(data=True):
        new_graph.add_edge(nodes_map[src], nodes_map[dst], **data)

    return new_graph


def add_labels(graph: networkx.DiGraph):
    new_graph = networkx.DiGraph()
    nodes_map = {}
    for node in graph:
        lbl = ailment.Stmt.Label(None, f"LABEL_{node.addr:x}", ins_addr=node.addr, block_idx=node.idx)
        node_copy = node.copy()
        node_copy.statements = [lbl, *node_copy.statements]
        nodes_map[node] = node_copy

    for old_node in graph.nodes:
        new_graph.add_node(nodes_map[old_node])

    for src, dst in graph.edges:
        new_graph.add_edge(nodes_map[src], nodes_map[dst])

    return new_graph


def update_labels(graph: networkx.DiGraph):
    """
    A utility function to recreate the labels for every node in an AIL graph. This useful when you are working with
    a graph where only _some_ of the nodes have labels.
    """
    return add_labels(remove_labels(graph))


def _flatten_structured_node(packed_node: SequenceNode | MultiNode) -> list[ailment.Block]:
    if not packed_node or not packed_node.nodes:
        return []

    blocks = []
    if packed_node.nodes is not None:
        for _node in packed_node.nodes:
            if isinstance(_node, (SequenceNode, MultiNode)):
                blocks += _flatten_structured_node(_node)
            else:
                blocks.append(_node)

    return blocks


def _find_node_in_graph(node: ailment.Block, graph: networkx.DiGraph) -> ailment.Block | None:
    for bb in graph:
        if isinstance(bb, ailment.Block) and bb.addr == node.addr and bb.idx == node.idx:
            return bb
    return None


def structured_node_has_multi_predecessors(
    node: SequenceNode | MultiNode | ailment.Block, graph: networkx.DiGraph
) -> bool:
    if graph is None:
        return False

    first_block = None
    if isinstance(node, (SequenceNode, MultiNode)) and node.nodes:
        flat_blocks = _flatten_structured_node(node)
        node = flat_blocks[0]

    if isinstance(node, ailment.Block):
        first_block = node

    if first_block is not None:
        graph_node = _find_node_in_graph(first_block, graph)
        if graph_node is not None:
            return len(list(graph.predecessors(graph_node))) > 1

    return False


def structured_node_is_simple_return(
    node: SequenceNode | MultiNode, graph: networkx.DiGraph, use_packed_successors=False
) -> bool:
    """
    Will check if a "simple return" is contained within the node a simple returns looks like this:
    if (cond) {
      // simple return
      ...
      return 0;
    }
    ...

    Returns true on any block ending in linear statements and a return.
    """
    # sanity check: we need a graph to understand returning blocks
    if graph is None:
        return False

    last_block = None
    if isinstance(node, (SequenceNode, MultiNode)) and node.nodes:
        flat_blocks = _flatten_structured_node(node)
        if all(isinstance(block, ailment.Block) for block in flat_blocks):
            last_block = flat_blocks[-1]
    elif isinstance(node, ailment.Block):
        last_block = node

    valid_last_stmt = last_block is not None
    if last_block is not None and last_block.statements:
        valid_last_stmt = not isinstance(last_block.statements[-1], (ailment.Stmt.ConditionalJump, ailment.Stmt.Jump))

    if use_packed_successors:
        last_block = node

    if valid_last_stmt:
        # note that the block may not be the same block in the AIL graph post dephication. we must find the block again
        # in the graph.
        assert isinstance(last_block, ailment.Block)
        last_graph_block = _find_node_in_graph(last_block, graph)
        if last_graph_block is not None:
            succs = list(graph.successors(last_graph_block))
            return not succs or succs == [last_graph_block]
    return False


def structured_node_is_simple_return_strict(node: BaseNode | SequenceNode | MultiNode | ailment.Block) -> bool:
    """
    Returns True iff the node exclusively contains a return statement.
    """
    if isinstance(node, (SequenceNode, MultiNode)) and node.nodes:
        flat_blocks = _flatten_structured_node(node)
        if len(flat_blocks) != 1:
            return False
        node = flat_blocks[-1]

    return (
        isinstance(node, ailment.Block)
        and len(node.statements) == 1
        and isinstance(node.statements[0], ailment.Stmt.Return)
    )


def is_statement_terminating(stmt: ailment.statement.Statement, functions) -> bool:
    if isinstance(stmt, ailment.Stmt.Return):
        return True
    if isinstance(stmt, ailment.Stmt.SideEffectStatement) and isinstance(stmt.expr.target, ailment.Expr.Const):
        # is it calling a non-returning function?
        target_func_addr = stmt.expr.target.value
        try:
            func = functions.get_by_addr(target_func_addr)
            return func.returning is False
        except KeyError:
            pass
    return False


class _PeepholeExprsWalker(ailment.AILBlockRewriter):
    """
    Walker to apply peephole optimizers (1)
    """

    def __init__(self, *args, expr_opts: list[PeepholeOptimizationExprBase], **kwargs):
        self.expr_opts = expr_opts
        self.any_update = False

        super().__init__(*args, **kwargs)

    def _handle_expr(
        self, expr_idx: int, expr: ailment.Expr.Expression, stmt_idx: int, stmt: ailment.Stmt.Statement | None, block
    ) -> ailment.Expression:
        # process the expr
        expr = super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)
        old_expr = expr

        redo = True
        while redo:
            redo = False
            for expr_opt in self.expr_opts:
                if isinstance(expr, expr_opt.expr_classes):
                    r = expr_opt.optimize(expr, stmt_idx=stmt_idx, block=block)
                    if r is not None and r is not expr:
                        assert expr.bits == r.bits
                        expr = r
                        redo = True
                        break

        if expr is not old_expr:
            self.any_update = True

        return expr


def peephole_optimize_exprs(block, expr_opts):
    # run expression optimizers
    walker = _PeepholeExprsWalker(expr_opts=expr_opts)
    walker.walk(block)
    return walker.any_update


class _PeepholeExprWalker(ailment.AILBlockRewriter):
    """
    Walker to apply peephole optimizers (2)
    """

    def __init__(self, *args, expr_opts: list[PeepholeOptimizationExprBase], **kwargs):
        self.expr_opts = expr_opts

        super().__init__(*args, **kwargs)

    def _handle_expr(
        self, expr_idx: int, expr: ailment.Expr.Expression, stmt_idx: int, stmt: ailment.Stmt.Statement | None, block
    ) -> ailment.Expression:
        redo = True
        while redo:
            redo = False
            for expr_opt in self.expr_opts:
                if isinstance(expr, expr_opt.expr_classes):
                    r = expr_opt.optimize(expr)
                    if r is not None and r is not expr:
                        expr = r
                        redo = True
                        break

        # continue to process the expr
        return super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)


def peephole_optimize_expr(expr: ailment.Expression, expr_opts: list[PeepholeOptimizationExprBase]):
    # run expression optimizers
    walker = _PeepholeExprWalker(expr_opts=expr_opts)
    return walker.walk_expression(expr, 0, None, None)


def copy_graph(graph: networkx.DiGraph[Block]) -> networkx.DiGraph[Block]:
    """
    Copy AIL Graph.

    :return: A copy of the AIl graph.
    """
    graph_copy = networkx.DiGraph()
    block_mapping = {}
    # copy all blocks
    for block in graph.nodes():
        new_block = copy.copy(block)
        new_stmts = copy.copy(block.statements)
        new_block.statements = new_stmts
        block_mapping[block] = new_block
        graph_copy.add_node(new_block)

    # copy all edges
    for src, dst, data in graph.edges(data=True):
        new_src = block_mapping[src]
        new_dst = block_mapping[dst]
        graph_copy.add_edge(new_src, new_dst, **data)
    return graph_copy


def peephole_optimize_stmts(block, stmt_opts):
    any_update = False
    statements = []

    # run statement optimizers
    # note that an optimizer may optionally edit or remove statements whose statement IDs are greater than stmt_idx
    stmt_idx = 0
    while stmt_idx < len(block.statements):
        stmt = block.statements[stmt_idx]
        old_stmt = stmt
        redo = True
        while redo:
            redo = False
            for opt in stmt_opts:
                if isinstance(stmt, opt.stmt_classes):
                    r = opt.optimize(stmt, stmt_idx=stmt_idx, block=block)
                    if r is not None and r is not stmt:
                        stmt = r
                        if r == ():
                            # the statement is gone; no more redo
                            redo = False
                            break
                        redo = True
                        break

        if stmt is not None and stmt is not old_stmt:
            if stmt != ():
                statements.append(stmt)
            any_update = True
        else:
            statements.append(old_stmt)
        stmt_idx += 1

    return statements, any_update


def match_stmt_classes(all_stmts: list, idx: int, stmt_class_seq: Iterable[type]) -> bool:
    for i, cls in enumerate(stmt_class_seq):
        if idx + i >= len(all_stmts):
            return False
        if not isinstance(all_stmts[idx + i], cls):
            return False
    return True


def peephole_optimize_multistmts(block, stmt_opts: list[PeepholeOptimizationMultiStmtBase]):
    any_update = False
    statements = block.statements[::]

    # run multi-statement optimizers
    stmt_idx = 0
    while stmt_idx < len(statements):
        redo = True
        while redo and stmt_idx < len(statements):
            redo = False
            for opt in stmt_opts:
                matched = False
                stmt_seq_len = None
                for stmt_class_seq_or_method in opt.stmt_classes:
                    if isinstance(stmt_class_seq_or_method, FunctionType):
                        r = stmt_class_seq_or_method(statements, stmt_idx)
                        if r > 0:
                            stmt_seq_len = r
                            matched = True
                            break
                    elif match_stmt_classes(statements, stmt_idx, stmt_class_seq_or_method):
                        stmt_seq_len = len(stmt_class_seq_or_method)
                        matched = True
                        break

                if matched:
                    assert stmt_seq_len is not None
                    matched_stmts = statements[stmt_idx : stmt_idx + stmt_seq_len]
                    r = opt.optimize(matched_stmts, stmt_idx=stmt_idx, block=block)
                    if r is not None:
                        # update statements
                        statements = statements[:stmt_idx] + r + statements[stmt_idx + stmt_seq_len :]
                        any_update = True
                        redo = True
                        break

        # move on to the next statement
        stmt_idx += 1

    return statements, any_update


def decompile_functions(
    path,
    functions: list[int | str] | None = None,
    structurer: str | None = None,
    catch_errors: bool = False,
    show_casts: bool = True,
    base_address: int | None = None,
    preset: str | None = None,
    cca: bool = False,
    cca_callsites: bool = True,
    llm: bool = False,
    progressbar: bool = False,
) -> str:
    """
    Decompile a binary into a set of functions.

    :param path:            The path to the binary to decompile.
    :param functions:       The functions to decompile. If None, all functions will be decompiled.
    :param structurer:      The structuring algorithms to use.
    :param catch_errors:    The structuring algorithms to use.
    :param show_casts:      Whether to show casts in the decompiled output.
    :param base_address:    The base address of the binary.
    :param preset:          The configuration preset to use during decompilation.
    :return:                The decompilation of all functions appended in order.
    """
    # delayed imports to avoid circular imports
    from angr.analyses.decompiler.decompilation_options import PARAM_TO_OPTION
    from angr.analyses.decompiler.structuring import DEFAULT_STRUCTURER

    structurer = structurer or DEFAULT_STRUCTURER.NAME

    path = pathlib.Path(path).resolve().absolute()
    # resolve loader args
    loader_main_opts_kwargs = {}
    if base_address is not None:
        loader_main_opts_kwargs["base_addr"] = base_address
    proj = angr.Project(path, auto_load_libs=False, main_opts=loader_main_opts_kwargs)
    cfg = proj.analyses.CFG(normalize=True, data_references=True, show_progressbar=progressbar)
    if cca:
        proj.analyses.CompleteCallingConventions(analyze_callsites=cca_callsites, show_progressbar=progressbar)

    # collect all functions when None are provided
    if functions is None:
        functions = sorted(cfg.kb.functions)

    # normalize the functions that could be ints as names
    normalized_functions: list[int | str] = []
    for func in functions:
        try:
            normalized_name = int(func, 0) if isinstance(func, str) else func
        except ValueError:
            normalized_name = func
        normalized_functions.append(normalized_name)
    functions = normalized_functions

    # verify that all functions exist
    for func in list(functions):
        if func not in cfg.functions:
            if catch_errors:
                _l.warning("Function %s does not exist in the CFG.", str(func))
                functions.remove(func)
            else:
                raise ValueError(f"Function {func} does not exist in the CFG.")

    # decompile all functions
    decompilation = ""
    dec_options = [
        (PARAM_TO_OPTION["structurer_cls"], structurer),
        (PARAM_TO_OPTION["show_casts"], show_casts),
    ]
    if llm:
        dec_options.append(("llm_refine", True))
    for func in functions:
        f = cfg.functions[func]
        if f is None or f.is_plt or f.is_syscall or f.is_alignment or f.is_simprocedure:
            continue

        exception_string = ""
        if not catch_errors:
            dec = proj.analyses.Decompiler(f, cfg=cfg, options=dec_options, preset=preset, show_progressbar=progressbar)
        else:
            try:
                # TODO: add a timeout
                dec = proj.analyses.Decompiler(
                    f, cfg=cfg, options=dec_options, preset=preset, show_progressbar=progressbar, fail_fast=True
                )
            except Exception as e:
                exception_string = str(e).replace("\n", " ")
                dec = None

        # do sanity checks on decompilation, skip checks if we already errored
        if not exception_string and (dec is None or not dec.codegen or not dec.codegen.text):
            exception_string = "Decompilation had no code output (failed in decompilation)"

        if exception_string:
            _l.critical("Failed to decompile %s because %s", repr(f), exception_string)
            decompilation += f"// [error: {func} | {exception_string}]\n"
        else:
            if dec is not None and dec.codegen is not None and dec.codegen.text is not None:
                decompilation += dec.codegen.text
            else:
                decompilation += "Invalid decompilation output"
            decompilation += "\n"

    return decompilation


def calls_in_graph(graph: networkx.DiGraph, consider_conditions: bool = False) -> int:
    """
    Counts the number of calls in a graph full of AIL Blocks
    """
    counter = AILBlockCallCounter(consider_conditions=consider_conditions)
    for node in graph.nodes:
        counter.walk(node)

    return counter.calls


def call_stmts_in_graph(
    graph: networkx.DiGraph, consider_conditions: bool = False
) -> tuple[
    list[tuple[tuple[Address, int], ailment.Stmt.SideEffectStatement]],
    list[tuple[tuple[Address, int], ailment.Expr.Call]],
]:
    """
    Return lists of call statements and call expressions in a given AIL graph.
    """
    counter = AILBlockCallCounter(consider_conditions=consider_conditions)
    for node in graph.nodes:
        counter.walk(node)
    # the above has an interface which includes nullable addresses because block can be none
    # but we always specify block here so we can ignore the Nones
    return counter.call_stmts, counter.call_exprs  # type: ignore


def has_addr_dups(graph: networkx.DiGraph[Block]) -> bool:
    return len({block.addr for block in graph}) != len(graph)


def find_block_by_addr(graph: networkx.DiGraph, addr, insn_addr=False) -> ailment.Block:
    for block in graph.nodes():
        if insn_addr:
            for stmt in block.statements:
                if "ins_addr" in stmt.tags and stmt.tags["ins_addr"] == addr:
                    return block
        else:
            if block.addr == addr:
                return block

    raise ValueError("The block is not in the graph!")


def find_block_by_addr_and_idx(graph: networkx.DiGraph, addr: int, idx: int | None) -> ailment.Block:
    for block in graph.nodes():
        if block.addr == addr and block.idx == idx:
            return block

    raise ValueError("The block is not in the graph!")


def sequence_to_blocks(seq: BaseNode) -> list[ailment.Block]:
    """
    Converts a sequence node (BaseNode) to a list of ailment blocks contained in it and all its children.
    """
    walker = SequenceToBlocks()
    walker.walk(seq)
    return walker.blocks


def sequence_to_statements(
    seq: BaseNode, exclude=(ailment.statement.Jump, ailment.statement.Jump)
) -> list[ailment.statement.Statement]:
    """
    Converts a sequence node (BaseNode) to a list of ailment Statements contained in it and all its children.
    May exclude certain types of statements.
    """
    statements = []
    blocks = sequence_to_blocks(seq)
    block: ailment.Block
    for block in blocks:
        if not block.statements:
            continue

        for stmt in block.statements:
            if isinstance(stmt, exclude):
                continue
            statements.append(stmt)

    return statements


def remove_edges_in_ailgraph(
    ail_graph: networkx.DiGraph, edges_to_remove: list[tuple[tuple[int, int | None], tuple[int, int | None]]]
) -> None:
    d = {(bb.addr, bb.idx): bb for bb in ail_graph}
    for src_addr, dst_addr in edges_to_remove:
        if src_addr in d and dst_addr in d and ail_graph.has_edge(d[src_addr], d[dst_addr]):
            ail_graph.remove_edge(d[src_addr], d[dst_addr])


# delayed import
from .structuring.structurer_nodes import (
    MultiNode,
    BaseNode,
    CodeNode,
    SequenceNode,
    ConditionNode,
    SwitchCaseNode,
    CascadingConditionNode,
    LoopNode,
)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: angr/protos/cfg.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'angr/protos/cfg.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from angr.protos import primitives_pb2 as angr_dot_protos_dot_primitives__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x61ngr/protos/cfg.proto\x12\x0b\x61ngr.protos\x1a\x1c\x61ngr/protos/primitives.proto\"]\n\x07\x43\x46GNode\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\r\x12\x10\n\x08\x62lock_id\x18\x03 \x03(\x03\x12\x11\n\treturning\x18\x04 \x01(\x08\x12\x13\n\x0binstr_addrs\x18\x05 \x03(\x04\"\xcd\x01\n\x03\x43\x46G\x12\r\n\x05ident\x18\x01 \x01(\t\x12#\n\x05nodes\x18\x02 \x03(\x0b\x32\x14.angr.protos.CFGNode\x12 \n\x05\x65\x64ges\x18\x03 \x03(\x0b\x32\x11.angr.protos.Edge\x12,\n\x0bmemory_data\x18\x04 \x03(\x0b\x32\x17.angr.protos.MemoryData\x12\x12\n\nnormalized\x18\x05 \x01(\x08\x12.\n\x0bjump_tables\x18\x06 \x03(\x0b\x32\x19.angr.protos.IndirectJump\"\xfb\x02\n\nMemoryData\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x11\n\x04size\x18\x02 \x01(\rH\x00\x88\x01\x01\x12\x34\n\x04type\x18\x03 \x01(\x0e\x32&.angr.protos.MemoryData.MemoryDataType\x12\x1b\n\x0ereference_size\x18\x04 \x01(\rH\x01\x88\x01\x01\"\xde\x01\n\x0eMemoryDataType\x12\x13\n\x0fUnknownDataType\x10\x00\x12\x0f\n\x0bUnspecified\x10\x01\x12\x0b\n\x07Integer\x10\x02\x12\x10\n\x0cPointerArray\x10\x03\x12\n\n\x06String\x10\x04\x12\x11\n\rUnicodeString\x10\x05\x12\x13\n\x0fSegmentBoundary\x10\x06\x12\x11\n\rCodeReference\x10\x07\x12\x0f\n\x0bGOTPLTEntry\x10\x08\x12\r\n\tELFHeader\x10\t\x12\x11\n\rFloatingPoint\x10\n\x12\r\n\tAlignment\x10\x0b\x42\x07\n\x05_sizeB\x11\n\x0f_reference_size\"^\n\rJumptableInfo\x12\x11\n\x04\x61\x64\x64r\x18\x01 \x01(\x04H\x00\x88\x01\x01\x12\x0c\n\x04size\x18\x02 \x01(\x05\x12\x12\n\nentry_size\x18\x03 \x01(\x05\x12\x0f\n\x07\x65ntries\x18\x04 \x03(\x04\x42\x07\n\x05_addr\"\xd0\x01\n\x0cIndirectJump\x12\x0c\n\x04\x61\x64\x64r\x18\x01 \x01(\x04\x12\x10\n\x08ins_addr\x18\x02 \x01(\x04\x12\x11\n\tfunc_addr\x18\x03 \x01(\x04\x12\x10\n\x08jumpkind\x18\x04 \x01(\t\x12\x10\n\x08stmt_idx\x18\x05 \x01(\x05\x12\x18\n\x10resolved_targets\x18\x06 \x03(\x04\x12\x11\n\tjumptable\x18\x07 \x01(\x08\x12.\n\njumptables\x18\x08 \x03(\x0b\x32\x1a.angr.protos.JumptableInfo\x12\x0c\n\x04type\x18\t \x01(\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'angr.protos.cfg_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CFGNODE']._serialized_start=68
  _globals['_CFGNODE']._serialized_end=161
  _globals['_CFG']._serialized_start=164
  _globals['_CFG']._serialized_end=369
  _globals['_MEMORYDATA']._serialized_start=372
  _globals['_MEMORYDATA']._serialized_end=751
  _globals['_MEMORYDATA_MEMORYDATATYPE']._serialized_start=501
  _globals['_MEMORYDATA_MEMORYDATATYPE']._serialized_end=723
  _globals['_JUMPTABLEINFO']._serialized_start=753
  _globals['_JUMPTABLEINFO']._serialized_end=847
  _globals['_INDIRECTJUMP']._serialized_start=850
  _globals['_INDIRECTJUMP']._serialized_end=1058
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import os.path

import angr
from angr.rustylib.fuzzer import Fuzzer, InMemoryCorpus, OnDiskCorpus

from tests.common import bin_location

# pylint: disable=missing-class-docstring,no-self-use


class TestFuzzer:
    def test_fuzzer(self):
        project = angr.Project(
            os.path.join(bin_location, "tests", "x86_64", "fauxware"), auto_load_libs=True, use_sim_procedures=False
        )
        stdin = angr.SimFile(content=b"", concrete=True)
        base_state = project.factory.entry_state(stdin=stdin, add_options={angr.options.STRICT_PAGE_ACCESS})

        def apply_fn(state: angr.SimState, input: bytes):  # pylint: disable=redefined-builtin
            print("Apply function called with input:", input, flush=True)
            assert state.project is not None
            cc = state.project.factory.cc()
            assert cc.return_addr is not None
            cc.return_addr.set_value(state, 0xDEADBEEF)
            state.posix.stdin.write(0, b"usernameN")
            state.posix.stdin.write(0, input)
            print("Done")

        corpus = InMemoryCorpus.from_list([b"S", b"SO", b"SOS", b"SOSN"])
        solutions = InMemoryCorpus()

        fuzzer = Fuzzer(base_state, corpus, solutions, apply_fn, 0, 0)

        new_corpus_entry = fuzzer.run_once()
        print("New corpus entry:", new_corpus_entry, flush=True)
        live_corpus = fuzzer.corpus()
        assert isinstance(live_corpus, InMemoryCorpus)
        assert 0 <= new_corpus_entry < len(live_corpus)
        print("Value: ", live_corpus[new_corpus_entry], flush=True)
        print("Fuzzer run completed", flush=True)

        new_corpus_entry = fuzzer.run_once()
        print("New corpus entry:", new_corpus_entry, flush=True)
        live_corpus = fuzzer.corpus()
        assert isinstance(live_corpus, InMemoryCorpus)
        assert 0 <= new_corpus_entry < len(live_corpus)
        print("Value: ", live_corpus[new_corpus_entry], flush=True)
        print("Fuzzer run completed", flush=True)

    def test_fuzzer_ondisk(self, tmp_path):
        print("Running test: ondisk", flush=True)
        project = angr.Project(
            os.path.join(bin_location, "tests", "x86_64", "fauxware"), auto_load_libs=True, use_sim_procedures=False
        )
        stdin = angr.SimFile(content=b"", concrete=True)
        base_state = project.factory.entry_state(stdin=stdin, add_options={angr.options.STRICT_PAGE_ACCESS})

        def apply_fn(state: angr.SimState, input: bytes):  # pylint: disable=redefined-builtin
            print("Apply function called with input:", input, flush=True)
            assert state.project is not None
            cc = state.project.factory.cc()
            assert cc.return_addr is not None
            cc.return_addr.set_value(state, 0xDEADBEEF)
            state.posix.stdin.write(0, b"usernameN")
            state.posix.stdin.write(0, input)
            print("Done")

        corpus_dir = str(tmp_path / "corpus")
        os.makedirs(corpus_dir, exist_ok=True)
        corpus = OnDiskCorpus(corpus_dir)
        for seed in [b"S", b"SO", b"SOS", b"SOSN"]:
            corpus.add(seed)
        solutions_dir = str(tmp_path / "solutions")
        os.makedirs(solutions_dir, exist_ok=True)
        solutions = OnDiskCorpus(solutions_dir)

        fuzzer = Fuzzer(base_state, corpus, solutions, apply_fn, 0, 0)

        new_corpus_entry = fuzzer.run_once()
        print("New corpus entry:", new_corpus_entry, flush=True)
        live_corpus = fuzzer.corpus()
        assert isinstance(live_corpus, OnDiskCorpus)
        assert 0 <= new_corpus_entry < len(live_corpus)
        print("Value: ", live_corpus[new_corpus_entry], flush=True)
        print("Fuzzer run completed", flush=True)
        live_solutions = fuzzer.solutions()
        assert isinstance(live_solutions, OnDiskCorpus)

    def test_fuzzer_mixed_inmem_corpus_ondisk_solutions(self, tmp_path):
        print("Running test: mixed_inmem_corpus_ondisk_solutions", flush=True)
        project = angr.Project(
            os.path.join(bin_location, "tests", "x86_64", "fauxware"), auto_load_libs=True, use_sim_procedures=False
        )
        stdin = angr.SimFile(content=b"", concrete=True)
        base_state = project.factory.entry_state(stdin=stdin, add_options={angr.options.STRICT_PAGE_ACCESS})

        def apply_fn(state: angr.SimState, input: bytes):  # pylint: disable=redefined-builtin
            print("Apply function called with input:", input, flush=True)
            assert state.project is not None
            cc = state.project.factory.cc()
            assert cc.return_addr is not None
            cc.return_addr.set_value(state, 0xDEADBEEF)
            state.posix.stdin.write(0, b"usernameN")
            state.posix.stdin.write(0, input)
            print("Done")

        corpus = InMemoryCorpus.from_list([b"S", b"SO", b"SOS", b"SOSN"])
        solutions_dir = str(tmp_path / "solutions_mixed1")
        os.makedirs(solutions_dir, exist_ok=True)
        solutions = OnDiskCorpus(solutions_dir)

        fuzzer = Fuzzer(base_state, corpus, solutions, apply_fn, 0, 0)
        new_corpus_entry = fuzzer.run_once()
        print("New corpus entry:", new_corpus_entry, flush=True)
        live_corpus = fuzzer.corpus()
        assert isinstance(live_corpus, InMemoryCorpus)
        assert 0 <= new_corpus_entry < len(live_corpus)
        print("Value: ", live_corpus[new_corpus_entry], flush=True)
        print("Fuzzer run completed", flush=True)
        live_solutions = fuzzer.solutions()
        assert isinstance(live_solutions, OnDiskCorpus)

    def test_fuzzer_mixed_ondisk_corpus_inmem_solutions(self, tmp_path):
        print("Running test: mixed_ondisk_corpus_inmem_solutions", flush=True)
        project = angr.Project(
            os.path.join(bin_location, "tests", "x86_64", "fauxware"), auto_load_libs=True, use_sim_procedures=False
        )
        stdin = angr.SimFile(content=b"", concrete=True)
        base_state = project.factory.entry_state(stdin=stdin, add_options={angr.options.STRICT_PAGE_ACCESS})

        def apply_fn(state: angr.SimState, input: bytes):  # pylint: disable=redefined-builtin
            print("Apply function called with input:", input, flush=True)
            assert state.project is not None
            cc = state.project.factory.cc()
            assert cc.return_addr is not None
            cc.return_addr.set_value(state, 0xDEADBEEF)
            state.posix.stdin.write(0, b"usernameN")
            state.posix.stdin.write(0, input)
            print("Done")

        corpus_dir = str(tmp_path / "corpus_mixed2")
        os.makedirs(corpus_dir, exist_ok=True)
        corpus = OnDiskCorpus(corpus_dir)
        for seed in [b"S", b"SO", b"SOS", b"SOSN"]:
            corpus.add(seed)
        solutions = InMemoryCorpus()

        fuzzer = Fuzzer(base_state, corpus, solutions, apply_fn, 0, 0)
        new_corpus_entry = fuzzer.run_once()
        print("New corpus entry:", new_corpus_entry, flush=True)
        live_corpus = fuzzer.corpus()
        assert isinstance(live_corpus, OnDiskCorpus)
        assert 0 <= new_corpus_entry < len(live_corpus)
        print("Value: ", live_corpus[new_corpus_entry], flush=True)
        print("Fuzzer run completed", flush=True)
        live_solutions = fuzzer.solutions()
        assert isinstance(live_solutions, InMemoryCorpus)



# pylint:disable=arguments-differ,too-many-boolean-expressions
from __future__ import annotations
import string

from archinfo import Endness

from angr.ailment.expression import Call, Const, StackBaseOffset, VirtualVariable, UnaryOp
from angr.ailment.statement import Assignment, Store, Statement, SideEffectStatement

from angr import SIM_LIBRARIES
from angr.utils.endness import ail_const_to_be
from .base import PeepholeOptimizationStmtBase

ASCII_PRINTABLES = set(string.printable)
ASCII_DIGITS = set(string.digits)


class InlinedStrcpy(PeepholeOptimizationStmtBase):
    """
    Simplifies inlined string copying logic into calls to strcpy.
    """

    __slots__ = ()

    NAME = "Simplifying inlined strcpy"
    stmt_classes = (Assignment, Store)

    def optimize(self, stmt: Assignment | Store, stmt_idx: int | None = None, block=None, **kwargs):
        inlined_strcpy_candidate = False
        src: Const | None = None
        strcpy_dst: StackBaseOffset | UnaryOp | None = None

        assert self.project is not None

        if (
            isinstance(stmt, Assignment)
            and isinstance(stmt.dst, VirtualVariable)
            and stmt.dst.was_stack
            and isinstance(stmt.src, Const)
            and isinstance(stmt.src.value, int)
        ):
            inlined_strcpy_candidate = True
            src = stmt.src
            strcpy_dst = StackBaseOffset(self.manager.next_atom(), self.project.arch.bits, stmt.dst.stack_offset)
        elif (
            isinstance(stmt, Store)
            and isinstance(stmt.addr, UnaryOp)
            and stmt.addr.op == "Reference"
            and isinstance(stmt.addr.operand, VirtualVariable)
            and stmt.addr.operand.was_stack
            and isinstance(stmt.data, Const)
            and isinstance(stmt.data.value, int)
        ) or (
            isinstance(stmt, Store)
            and isinstance(stmt.addr, StackBaseOffset)
            and isinstance(stmt.data, Const)
            and isinstance(stmt.data.value, int)
        ):
            inlined_strcpy_candidate = True
            src = stmt.data
            strcpy_dst = stmt.addr

        if inlined_strcpy_candidate:
            assert src is not None and strcpy_dst is not None
            assert isinstance(src.value, int)
            assert self.kb is not None

            r, s = self.is_integer_likely_a_string(src.value, src.size, self.project.arch.memory_endness)
            if r:
                assert s is not None

                # replace it with a call to strncpy
                str_id = self.kb.custom_strings.allocate(s.encode("ascii"))
                return SideEffectStatement(
                    stmt.idx,
                    Call(
                        stmt.idx,
                        "strncpy",
                        args=[
                            strcpy_dst,
                            Const(None, None, str_id, self.project.arch.bits, custom_string=True),
                            Const(None, None, len(s), self.project.arch.bits),
                        ],
                        prototype=SIM_LIBRARIES["libc.so"][0].get_prototype("strncpy", arch=self.project.arch),
                        **stmt.tags,
                    ),
                    **stmt.tags,
                )

            # scan forward in the current block to find all consecutive constant stores
            if block is not None and stmt_idx is not None:
                all_constant_stores: dict[int, tuple[int, Const | None]] = self.collect_constant_stores(block, stmt_idx)
                if all_constant_stores:
                    offsets = sorted(all_constant_stores.keys())
                    next_offset = min(offsets)
                    stride = []
                    for offset in offsets:
                        if next_offset is not None and offset != next_offset:
                            next_offset = None
                            stride = []
                        stmt_idx_, v = all_constant_stores[offset]
                        if v is not None:
                            stride.append((offset, stmt_idx_, v))
                            next_offset = offset + v.size
                        else:
                            next_offset = None
                            stride = []

                    if not stride:
                        return None
                    min_stride_stmt_idx = min(stmt_idx_ for _, stmt_idx_, _ in stride)
                    if min_stride_stmt_idx > stmt_idx:
                        # the current statement is not involved in the stride. we can't simplify here, otherwise we
                        # will incorrectly remove the current statement
                        return None

                    integer, size = self.stride_to_int(stride)
                    prev_stmt = None if stmt_idx == 0 else block.statements[stmt_idx - 1]
                    min_str_length = 1 if prev_stmt is not None and self.is_inlined_strcpy(prev_stmt) else 4
                    r, s = self.is_integer_likely_a_string(integer, size, Endness.BE, min_length=min_str_length)
                    if r:
                        assert s is not None

                        # we remove all involved statements whose statement IDs are greater than the current one
                        for _, stmt_idx_, _ in reversed(stride):
                            if stmt_idx_ <= stmt_idx:
                                continue
                            block.statements[stmt_idx_] = None
                        block.statements = [ss for ss in block.statements if ss is not None]

                        str_id = self.kb.custom_strings.allocate(s.encode("ascii"))
                        return SideEffectStatement(
                            stmt.idx,
                            Call(
                                stmt.idx,
                                "strncpy",
                                args=[
                                    strcpy_dst,
                                    Const(None, None, str_id, self.project.arch.bits, custom_string=True),
                                    Const(None, None, len(s), self.project.arch.bits),
                                ],
                                prototype=SIM_LIBRARIES["libc.so"][0].get_prototype("strncpy", arch=self.project.arch),
                                **stmt.tags,
                            ),
                            **stmt.tags,
                        )

        return None

    @staticmethod
    def stride_to_int(stride: list[tuple[int, int, Const]]) -> tuple[int, int]:
        stride = sorted(stride, key=lambda x: x[0])
        n = 0
        size = 0
        for _, _, v in stride:
            size += v.size
            n <<= v.bits
            assert isinstance(v.value, int)
            n |= v.value
        return n, size

    def collect_constant_stores(self, block, starting_stmt_idx: int) -> dict[int, tuple[int, Const | None]]:
        assert self.project is not None

        r = {}
        for idx, stmt in enumerate(block.statements):
            if idx < starting_stmt_idx:
                continue
            if (
                isinstance(stmt, Assignment)
                and isinstance(stmt.dst, VirtualVariable)
                and stmt.dst.was_stack
                and isinstance(stmt.dst.stack_offset, int)
            ):
                if isinstance(stmt.src, Const):
                    r[stmt.dst.stack_offset] = idx, ail_const_to_be(stmt.src, self.project.arch.memory_endness)
                else:
                    r[stmt.dst.stack_offset] = idx, None
            elif isinstance(stmt, Store) and isinstance(stmt.addr, StackBaseOffset):
                if isinstance(stmt.data, Const):
                    r[stmt.addr.offset] = idx, ail_const_to_be(stmt.data, self.project.arch.memory_endness)
                else:
                    r[stmt.addr.offset] = idx, None
        return r

    @staticmethod
    def is_integer_likely_a_string(v: int, size: int, endness: Endness, min_length: int = 4) -> tuple[bool, str | None]:
        # we need at least four bytes of printable characters

        chars = []
        if endness == Endness.LE:
            while v != 0:
                byt = v & 0xFF
                if chr(byt) not in ASCII_PRINTABLES:
                    return False, None
                chars.append(chr(byt))
                v >>= 8

        elif endness == Endness.BE:
            first_non_zero = False
            for _ in range(size):
                byt = v & 0xFF
                v >>= 8
                if byt == 0:
                    if first_non_zero:
                        return False, None
                    continue
                first_non_zero = True  # this is the first non-zero byte
                if chr(byt) not in ASCII_PRINTABLES:
                    return False, None
                chars.append(chr(byt))
            chars.reverse()
        else:
            # unsupported endness
            return False, None

        if len(chars) >= min_length:
            if len(chars) <= 4 and all(ch in ASCII_DIGITS for ch in chars):
                return False, None
            return True, "".join(chars)
        return False, None

    @staticmethod
    def is_inlined_strcpy(stmt: Statement) -> bool:
        return (
            isinstance(stmt, SideEffectStatement)
            and isinstance(stmt.expr.target, str)
            and stmt.expr.target == "strncpy"
            and stmt.expr.args is not None
            and len(stmt.expr.args) == 3
            and isinstance(stmt.expr.args[1], Const)
            and "custom_string" in stmt.expr.args[1].tags
        )



from __future__ import annotations
import claripy

import angr


class isblank(angr.SimProcedure):
    # pylint: disable=arguments-differ, missing-class-docstring
    def run(self, c):
        is_space = c == 32
        is_tab = c == 9
        int_size = self.arch.sizeof["int"]  # type: ignore[reportOptionalMemberAccess]

        return claripy.If(
            claripy.Or(is_space, is_tab),
            claripy.BVV(1, int_size),
            claripy.BVV(0, int_size),
        )
